{}//Unit under construction
{*************************************************************************}
{ GH Freebrary                                                            }
{                                                                         }
{ GHFCustomDataSetRec.pas - TghCustomDataSetRec class unit.               }
{                                                                         }
{ This software is distributed under the open source BSD license.         }
{                                                                         }
{ Copyright (c) 2002-2014, Al Gonzalez (rescatandoadelphi.blogspot.com)   }
{ All rights reserved.                                                    }
{                                                                         }
{ Redistribution and use in source and binary forms, with or without      }
{ modification, are permitted provided that the following conditions are  }
{ met:                                                                    }
{                                                                         }
{   - Redistributions of source code must retain the above copyright      }
{     notice, this list of conditions and the following disclaimer.       }
{   - Redistributions in binary form must reproduce the above copyright   }
{     notice, this list of conditions and the following disclaimer in the }
{     documentation and/or other materials provided with the              }
{     distribution.                                                       }
{   - Neither the name of Al Gonzalez nor the names of his contributors   }
{     may be used to endorse or promote products derived from this        }
{     software without specific prior written permission.                 }
{                                                                         }
{ THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS }
{ IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED   }
{ TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A         }
{ PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT      }
{ HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,  }
{ SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT        }
{ LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,   }
{ DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY   }
{ THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT     }
{ (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE   }
{ OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.    }
{*************************************************************************}

{ Please contribute to improve this library.  I am not a good programmer,
  but good programmers know what Delphi is worth.

  A.G. }

Unit GHFCustomDataSetRec;  { Custom Data Set Record }

Interface

  Uses
    Classes, DB, GHFRTL, SysUtils;

  Type
    { Custom Data Set Record class }
    TghCustomDataSetRec = Class (TPersistent)
      Protected
        { Static methods }
        Class Function CleanFieldName (Const Value :String) :String;
        Function FieldCount (Const Fields :String) :Integer;
        Function GetFieldBoolProp (Const Index :Integer) :Boolean;
        Function GetFieldFloatProp (Const Index :Integer) :Double;
        Function GetFieldIntProp (Const Index :Integer) :Integer;
        Function GetFieldProp (Const AccessMethod :Pointer;
          Const Index :Integer) :Variant; Overload;
        Function GetFieldProp (Const Index :Integer) :Variant; Overload;
        Function GetFieldRec (Var Fields :String;
          Out Rec :TghCustomDataSetRec;
          Const VerifyHasRefFields :Boolean = False) :Boolean;
        Function GetFieldStrProp (Const Index :Integer) :String;
        Function GetNextField (Const Fields :String;
          Var StartIndex :Integer; Var Rec :TghCustomDataSetRec;
          Out AField :String) :Boolean; Overload;
        Function GetNextField (Const Fields :String;
          Var StartIndex :Integer; Var Rec :TghCustomDataSetRec;
          Out AField :String; Var FieldIndex :Integer) :Boolean; Overload;
        Function GetPropName (Const AccessMethod :Pointer;
          Const AIndex :Integer) :String;
        Procedure RaiseError (Const Msg :String; Const Param :String = '');
        Procedure SetFieldBoolProp (Const Index :Integer;
          Const AValue :Boolean);
        Procedure SetFieldFloatProp (Const Index :Integer;
          Const AValue :Double);
        Procedure SetFieldIntProp (Const Index :Integer;
          Const AValue :Integer);
        Procedure SetFieldProp (Const AccessMethod :Pointer;
          Const Index :Integer; Const AValue :Variant); Overload;
        Procedure SetFieldProp (Const Index :Integer;
          Const AValue :Variant); Overload;
        Procedure SetFieldStrProp (Const Index :Integer;
          Const AValue :String);

        { Virtual methods }
        Function GetDataSet :TDataSet; Virtual; Abstract;
        Function GetValue (Const AField :String;
          IncludeFieldObjs :Boolean = True) :Variant; Virtual;
        Function GetValues (Const Fields :String = '') :Variant; Virtual;
        Function InternalField (Const Name :String) :TField; Virtual;
        Function InternalHasField (Const Name :String;
          IncludeFieldObjs :Boolean = True) :Boolean; Virtual;
        Function InternalRefRec (Const AField :String) :TghCustomDataSetRec;
          Virtual;
        Procedure SetValue (Const AField :String; Const AValue :Variant;
          IncludeFieldObjs :Boolean = True); Virtual;
        Procedure SetValues (Const Fields :String; Const AValues :Variant);
          Virtual;
      Public
        Destructor Destroy; Override;

        { Static methods }
        Function Add (Const AField :String; Const Addend :Double) :Variant;
          Overload;
        Function Add (Const AField, AddendField :String) :Variant;
          Overload;
        Function Add (Const Fields :String;
          Const CheckNulls :Boolean = False) :Variant; Overload;
        Procedure Assign (Const Rec :TghCustomDataSetRec;
          Const Fields :String); Reintroduce; Overload;
        Function Between (Const AValue :Double; Const MinField :String;
          Const MaxValue :Double) :Boolean; Overload;
        Function Between (Const AValue :Double;
          Const MinField, MaxField :String) :Boolean; Overload;
        Function Bool (Const Fields :String) :Boolean;
        Function Changed (Const Fields :String) :Boolean;
        Procedure Clear (Const Fields :String);
        Function Divide (Const AField :String; Const Divisor :Double;
          Const Decimals :Integer = MaxInt) :Variant; Overload;
        Function Divide (Const AField, DivisorField :String;
          Const Decimals :Integer = MaxInt) :Variant; Overload;
        Function Divide (Const Dividend :Double;
          Const DivisorField :String; Const Decimals :Integer = MaxInt)
          :Variant; Overload;
        Function Equals (Const Field1, Field2 :String) :Boolean;
        Function Field (Name :String) :TField;
        Function Float (Const Fields :String) :Double;
        Function Greater (Const Field1, Field2 :String) :Boolean;
        Function HasField (Name :String;
          Const IncludeFieldObjs :Boolean = True;
          Const IncludeRefs :Boolean = True) :Boolean;
        Function HasFieldValue (Const Name :String) :Boolean;
        Function HasValue (Const Fields :String) :Boolean;
        Procedure Increment (Const AField :String; Const Inc :Double = 1;
          Const CheckNulls :Boolean = False); Overload;
        Procedure Increment (Const AField, IncField :String;
          Const CheckNulls :Boolean = False); Overload;
        Function InRange (Const Value :Integer;
          Const MinField, MaxField :String) :Boolean; Overload;
        Function InRange (Const Field, MinField, MaxField :String)
          :Boolean; Overload;
        Function Int (Const Fields :String) :Integer;
        Function IntField (Const Name :String) :TIntegerField;
        Function IsGreater (Const Field1, Field2 :String) :Boolean;
        Function IsInRange (Const Value :Integer;
          Const MinField, MaxField :String) :Boolean; Overload;
        Function IsInRange (Const Field, MinField, MaxField :String)
          :Boolean; Overload;
        Function IsLess (Const Field1, Field2 :String) :Boolean;
        Function IsNull (Const Fields :String) :Boolean;
        Function Less (Const Field1, Field2 :String) :Boolean;
        Function Max (Const Fields :String) :Variant;
        Function Min (Const Fields :String) :Variant;
        Function Multiply (Const AField :String; Const Factor :Double;
          Const Decimals :Integer = MaxInt) :Variant; Overload;
        Function Multiply (Const AField, FactorField :String;
          Const Decimals :Integer = MaxInt) :Variant; Overload;
        Function Percent (Const AField :String; Const Percentage :Double;
          Const Decimals :Integer = MaxInt) :Variant; Overload;
        Function Percent (Const AField, PercentageField :String;
          Const Decimals :Integer = MaxInt) :Variant; Overload;
        Function RefRec (AField :String) :TghCustomDataSetRec;
        Procedure Revert (Const Fields :String);
        Function Str (Const Fields :String) :String;
        Function TwoInts (Const Field1, Field2 :String) :TghTwoInts;
          Overload;
        Function TwoInts (Const Fields :String) :TghTwoInts; Overload;
        Function Value (Const Fields :String) :Variant;

        { Virtual methods }
        Procedure Focus (Const AField :String); Virtual;

        { Properties }
        Property DataSet :TDataSet Read GetDataSet;
        Property Values [Const Fields :String] :Variant Read GetValues
          Write SetValues; Default;
    End;

    { Custom Data Set Record class type }
    TghCustomDataSetRecClass = Class Of TghCustomDataSetRec;

  Var
    GHDataSetRecFieldSigns :TSysCharSet = ['+', '-', '?', '@'];
    
Implementation

  Uses
    Variants, TypInfo, FmtBCD, GHFUtils, Math;

  { TghCustomDataSetRec }

  Destructor TghCustomDataSetRec.Destroy;
  Begin
    DataSet.Delete;
  End;

  { Protected static methods }

  Class Function TghCustomDataSetRec.CleanFieldName (Const Value :String)
    :String;
  Begin
    Result := ghDeleteLeft (Value, GHDataSetRecFieldSigns)
  End;

  Function TghCustomDataSetRec.FieldCount (Const Fields :String)
    :Integer;
  Begin
    Result := ghCount (Fields, [';', ',']) + 1;
  End;

  Function TghCustomDataSetRec.GetFieldBoolProp (Const Index :Integer)
    :Boolean;
  Begin
    Result := ghBool (GetFieldProp (
      @TghCustomDataSetRec.GetFieldBoolProp, Index));
  End;

  Function TghCustomDataSetRec.GetFieldFloatProp (Const Index :Integer)
    :Double;
  Begin
    Result := ghFloat (GetFieldProp (
      @TghCustomDataSetRec.GetFieldFloatProp, Index));
  End;

  Function TghCustomDataSetRec.GetFieldIntProp (Const Index :Integer)
    :Integer;
  Begin
    Result := ghInt (GetFieldProp (@TghCustomDataSetRec.GetFieldIntProp,
      Index));
  End;

  Function TghCustomDataSetRec.GetFieldProp (
    Const AccessMethod :Pointer; Const Index :Integer) :Variant;
  Begin
    Result := Values [GetPropName (AccessMethod, Index)];
  End;

  Function TghCustomDataSetRec.GetFieldProp (Const Index :Integer)
    :Variant;
  Var
    Method :Function (Const Index :Integer) :Variant Of Object;
  Begin
    Method := GetFieldProp;  { We avoid name ambiguity }
    Result := GetFieldProp (TMethod (Method).Code, Index);
  End;

  Function TghCustomDataSetRec.GetFieldRec (Var Fields :String;
    Out Rec :TghCustomDataSetRec;
    Const VerifyHasRefFields :Boolean = False) :Boolean;
  Begin
    Rec := Self;

    If (ghIndex (Fields, '.') = 0) Or HasField (Fields, True, False) Then
      Result := True
    Else
    Begin
      With ghLeftRight (Fields, '.') Do
        If Not VerifyHasRefFields Or InternalHasField (Value1) Then
        Begin
          Rec := InternalRefRec (Value1);

          If Rec <> Nil Then
          Begin
            Fields := Value2;
            Result := Rec.GetFieldRec (Fields, Rec, VerifyHasRefFields);
            Exit;
          End;
        End;

      Result := False;
    End;
  End;

  Function TghCustomDataSetRec.GetFieldStrProp (Const Index :Integer)
    :String;
  Begin
    Result := VarToStr (GetFieldProp (
      @TghCustomDataSetRec.GetFieldStrProp, Index));
  End;

  Function TghCustomDataSetRec.GetNextField (Const Fields :String;
    Var StartIndex :Integer; Var Rec :TghCustomDataSetRec;
    Out AField :String) :Boolean;
  Begin
    If (StartIndex = 1) Or (ghChr (Fields, StartIndex - 1) = ';') Then
      Rec := Self;

    While ghIsIndex (StartIndex, Fields) And
    Not (Fields [StartIndex] In [';', ',']) Do
    Begin
      AField := AField + Fields [StartIndex];
      Inc (StartIndex);
    End;

    Result := ghIsIndex (StartIndex - 1, Length (Fields) + 1);

    If Not Result Then
      Exit;

    If Rec <> Nil Then
      Rec.GetFieldRec (AField, Rec);

    Inc (StartIndex);
  End;

  Function TghCustomDataSetRec.GetNextField (Const Fields :String;
    Var StartIndex :Integer; Var Rec :TghCustomDataSetRec;
    Out AField :String; Var FieldIndex :Integer) :Boolean;
  Begin
    If StartIndex = 1 Then
      FieldIndex := -1;

    Result := GetNextField (Fields, StartIndex, Rec, AField);

    If Result Then
      Inc (FieldIndex);
  End;

  Function TghCustomDataSetRec.GetPropName (Const AccessMethod :Pointer;
    Const AIndex :Integer) :String;
  Var
    I :Integer;
    Infos :TghPPropInfoArr;
  Begin
    Infos := ghPropInfos (ClassType);

    For I := 0 To High (Infos) Do
      With Infos [I]^ Do
        If ((GetProc = AccessMethod) Or (SetProc = AccessMethod)) And
        (Index = AIndex) Then
        Begin
          Result := Name;
          Exit;
        End;

    RaiseError (Format ('Invalid property access method or index (%d).',
      [AIndex]));;
  End;

  Procedure TghCustomDataSetRec.RaiseError (Const Msg :String;
    Const Param :String = '');
  Begin
    DatabaseErrorFmt (Msg, [Param], DataSet);
  End;

  Procedure TghCustomDataSetRec.SetFieldBoolProp (Const Index :Integer;
    Const AValue :Boolean);
  Begin
    SetFieldProp (@TghCustomDataSetRec.SetFieldBoolProp, Index, AValue);
  End;

  Procedure TghCustomDataSetRec.SetFieldFloatProp (Const Index :Integer;
    Const AValue :Double);
  Begin
    SetFieldProp (@TghCustomDataSetRec.SetFieldFloatProp, Index,
      AValue);
  End;

  Procedure TghCustomDataSetRec.SetFieldIntProp (Const Index :Integer;
    Const AValue :Integer);
  Begin
    SetFieldProp (@TghCustomDataSetRec.SetFieldIntProp, Index, AValue);
  End;

  Procedure TghCustomDataSetRec.SetFieldProp (
    Const AccessMethod :Pointer; Const Index :Integer;
    Const AValue :Variant);
  Begin
    Values [GetPropName (AccessMethod, Index)] := AValue;
  End;

  Procedure TghCustomDataSetRec.SetFieldProp (Const Index :Integer;
    Const AValue :Variant);
  Var
    Method :Procedure (Const Index :Integer; Const AValue :Variant) Of
      Object;
  Begin
    Method := SetFieldProp;  { We avoid name ambiguity }
    SetFieldProp (TMethod (Method).Code, Index, AValue);
  End;
  
  Procedure TghCustomDataSetRec.SetFieldStrProp (Const Index :Integer;
    Const AValue :String);
  Begin
    SetFieldProp (@TghCustomDataSetRec.SetFieldStrProp, Index, AValue);
  End;

  { Protected virtual methods }

  Function TghCustomDataSetRec.GetValue (Const AField :String;
    IncludeFieldObjs :Boolean = True) :Variant;
  Var
    FieldName :String;
    Old :Boolean;
    Ptr :Pointer;
    Sign :Char;
  Begin
    FieldName := AField;
    
    // Sign prefix ("+" for absolute value, "-" for opposite sign value)
    Sign := ghChr (FieldName);

    If Sign In ['+', '-'] Then
      FieldName := ghCopy (FieldName, 2);

    // "?" prefix to get the old value of a data set field
    Old := ghChr (FieldName) = '?';

    If Old Then
      FieldName := ghCopy (FieldName, 2);

    If IncludeFieldObjs Then
    Begin
      Ptr := DataSet.FindField (FieldName);

      If Ptr <> Nil Then  // Data set field
        If Old Then
          Result := TField (Ptr).OldValue
        Else
          Result := TField (Ptr).Value///considerar Blobs, ver TBlobField.GetIsNull y TBlobField.GetAsString
      Else
        IncludeFieldObjs := False;
    End;

    If Not IncludeFieldObjs Then
    Begin
      Ptr := MethodAddress ('Get' + FieldName);

      If Ptr <> Nil Then  // Variant method of record object
        Result := TghVarMethod (ghMethodRec (Self, Ptr))
      Else
      Begin
        Ptr := GetPropInfo (Self, FieldName);

        If Ptr = Nil Then
          RaiseError (
            'Data set record does not have field, Get method or ' +
            'property "%s".', FieldName)
        Else
          If PPropInfo (Ptr).GetProc = Nil Then
            RaiseError ('Data set record property %s is write only.',
              FieldName)
          Else
            { We avoid recursive field properties }
            If ghPosPtr ([@TghCustomDataSetRec.GetFieldBoolProp,
            @TghCustomDataSetRec.GetFieldFloatProp,
            @TghCustomDataSetRec.GetFieldIntProp,
            @TghCustomDataSetRec.GetFieldProp,
            @TghCustomDataSetRec.GetFieldStrProp],
            PPropInfo (Ptr).GetProc) > -1
            Then
              RaiseError (
                'Data set record does not have field or Get method for ' +
                'index property %s.', FieldName)
            Else
              // Property of record object
              Result := GetPropValue (Self, FieldName, False);
      End;
    End;

    If Not VarIsNull (Result) Then
      Case Sign Of
        '+' : Result := Abs (Result);
        '-' :
        // The TFmtBCDVariantType class don't overrides UnaryOp method
        If VarType (Result) = VarFmtBCD Then
          Result := -Double (Result)
        Else
          Result := -Result;
      End;
  End;

  Function TghCustomDataSetRec.GetValues (Const Fields :String = '') :Variant;
  Var
    AField :String;
    I, J :Integer;
    Rec :TghCustomDataSetRec;
  Begin
    Result := ghVarArr (FieldCount (Fields), False);
    I := 1;

    While GetNextField (Fields, I, Rec, AField, J) Do
      If Rec <> Nil Then
        ghAssign (Result, J, Rec.GetValue (AField))
      Else
        ghAssign (Result, J, Null);
  End;

  Function TghCustomDataSetRec.InternalField (Const Name :String) :TField;
  Begin
    Result := DataSet.FieldByName (Name);
  End;

  Function TghCustomDataSetRec.InternalHasField (Const Name :String;
    IncludeFieldObjs :Boolean = True) :Boolean;
  Begin
    Result := (IncludeFieldObjs And (DataSet.FindField (Name) <> Nil)) Or
      (MethodAddress ('Get' + Name) <> Nil) Or (MethodAddress ('Set' +
      Name) <> Nil) Or IsPublishedProp (Self, Name);
  End;

  Function TghCustomDataSetRec.InternalRefRec (Const AField :String)
    :TghCustomDataSetRec;
  Begin
    Result := Nil;
  End;

  Procedure TghCustomDataSetRec.SetValue (Const AField :String;
    Const AValue :Variant; IncludeFieldObjs :Boolean = True);
  Var
    AFocus :Boolean;
    FieldName :String;
    Ptr :Pointer;
    SavedReadOnly :Boolean;
  Begin
    FieldName := AField;

    // "@" prefix to focus the field
    AFocus := ghChr (FieldName) = '@';

    If AFocus Then
      FieldName := ghCopy (FieldName, 2);

    If IncludeFieldObjs Then
    Begin
      Ptr := DataSet.FindField (FieldName);

      If Ptr <> Nil Then  // Data set field
      Begin
        If Not (DataSet.State In dsWriteModes) Then
          DataSet.Edit;

        // Field's ReadOnly property only for controls
        SavedReadOnly := TField (Ptr).ReadOnly;
        TField (Ptr).ReadOnly := False;

        Try
          DataSet [FieldName] := AValue;
        Finally
          TField (Ptr).ReadOnly := SavedReadOnly;
        End;
      End
      Else
        IncludeFieldObjs := False;
    End;

    If Not IncludeFieldObjs Then
    Begin
      Ptr := MethodAddress ('Set' + FieldName);

      If Ptr <> Nil Then  // Variant parameter method of record object
        TghVarParamMethod (ghMethodRec (Self, Ptr)) (AValue)
      Else
      Begin
        Ptr := GetPropInfo (Self, FieldName);

        If Ptr = Nil Then
          RaiseError (
            'Data set record does not have field, Set method or ' +
            'property "%s".', FieldName)
        Else
          If PPropInfo (Ptr).SetProc = Nil Then
            RaiseError ('Data set record property %s is read only.',
              FieldName)
          Else
            { We avoid recursive field properties }
            If ghPosPtr ([@TghCustomDataSetRec.SetFieldBoolProp,
            @TghCustomDataSetRec.SetFieldFloatProp,
            @TghCustomDataSetRec.SetFieldIntProp,
            @TghCustomDataSetRec.SetFieldProp,
            @TghCustomDataSetRec.SetFieldStrProp],
            PPropInfo (Ptr).SetProc) > -1
            Then
              RaiseError (
                'Data set record does not have field or Set method for ' +
                'index property %s.', FieldName)
            Else
              // Property of record object
              SetPropValue (Self, FieldName, AValue);
      End;
    End;

    If AFocus Then
      Focus (FieldName);
  End;

  Procedure TghCustomDataSetRec.SetValues (Const Fields :String;
    Const AValues :Variant);
  Var
    AField :String;
    I, J :Integer;
    Rec :TghCustomDataSetRec;
  Begin
    I := 1;

    While GetNextField (Fields, I, Rec, AField, J) Do
      If Rec <> Nil Then
        Rec.SetValue (AField, ghValue (AValues, J));
  End;

  { Public static methods }

  Function TghCustomDataSetRec.Add (Const AField :String;
    Const Addend :Double) :Variant;
  Begin
    If IsNull (AField) Then
      Result := Null
    Else
      Result := Value (AField) + Addend;
  End;

  Function TghCustomDataSetRec.Add (Const AField, AddendField :String)
    :Variant;
  Begin
    If IsNull (AddendField) Then
      Result := Null
    Else
      Result := Add (AField, Float (AddendField));
  End;

  Function TghCustomDataSetRec.Add (Const Fields :String;
    Const CheckNulls :Boolean = False) :Variant;
  Var
    AField :String;
    AValue :Variant;
    I :Integer;
    Rec :TghCustomDataSetRec;
  Begin
    Result := Null;
    I := 1;

    While GetNextField (Fields, I, Rec, AField) Do
      If Rec <> Nil Then
      Begin
        AValue := Rec.GetValue (AField);

        If Not VarIsNull (AValue) Then
          If VarIsNull (Result) Then
            Result := AValue
          Else
            Result := Result + AValue
        Else
          If CheckNulls Then
          Begin
            Result := Null;
            Exit;
          End
      End;

    If VarIsNull (Result) And Not CheckNulls Then
      Result := 0;
  End;

  Procedure TghCustomDataSetRec.Assign (
    Const Rec :TghCustomDataSetRec; Const Fields :String);
  Begin
    Values [Fields] := Rec [Fields];
  End;

  Function TghCustomDataSetRec.Between (Const AValue :Double;
    Const MinField :String; Const MaxValue :Double) :Boolean;
  Begin
    If IsNull (MinField) Then
      Result := False
    Else
      Result := (AValue >= Float (MinField)) And (AValue <= MaxValue);
  End;                                    

  Function TghCustomDataSetRec.Between (Const AValue :Double;
    Const MinField, MaxField :String) :Boolean;
  Begin
    If IsNull (MaxField) Then
      Result := False
    Else
      Result := Between (AValue, MinField, Float (MaxField));
  End;
  
  Function TghCustomDataSetRec.Bool (Const Fields :String) :Boolean;
  Begin
    Result := ghBool (Value (Fields));
  End;

  Function TghCustomDataSetRec.Changed (Const Fields :String) :Boolean;
  Var
    AField :String;
    I :Integer;
    Rec :TghCustomDataSetRec;
  Begin
    I := 1;

    While GetNextField (Fields, I, Rec, AField) Do
      If (Rec <> Nil) And Not ghEquals (Rec.GetValue (AField),
      Rec.GetValue ('?' + AField)) Then
      Begin
        Result := True;
        Exit;
      End;

    Result := False;
  End;

  Procedure TghCustomDataSetRec.Clear (Const Fields :String);
  Begin
    Values [Fields] := Null;
  End;

  Function TghCustomDataSetRec.Divide (Const AField :String;
    Const Divisor :Double; Const Decimals :Integer = MaxInt) :Variant;
  Begin
    If IsNull (AField) Then
      Result := Null
    Else
      Result := ghRound (Value (AField) / Divisor, Decimals);
  End;

  Function TghCustomDataSetRec.Divide (
    Const AField, DivisorField :String; Const Decimals :Integer = MaxInt)
    :Variant;
  Begin
    If IsNull (DivisorField) Then
      Result := Null
    Else
      Result := Divide (AField, Float (DivisorField), Decimals);
  End;

  Function TghCustomDataSetRec.Divide (Const Dividend :Double;
    Const DivisorField :String; Const Decimals :Integer = MaxInt) :Variant;
  Begin
    If IsNull (DivisorField) Then
      Result := Null
    Else
      Result := ghRound (Dividend / Value (DivisorField), Decimals);
  End;

  Function TghCustomDataSetRec.Equals (Const Field1, Field2 :String)
    :Boolean;
  Begin
    Result := Value (Field1) = Value (Field2);
  End;

  Function TghCustomDataSetRec.Field (Name :String) :TField;
  Var
    Rec :TghCustomDataSetRec;
  Begin
    If GetFieldRec (Name, Rec) Then
      Result := Rec.InternalField (Name)
    Else
      Result := Nil;
  End;

  Function TghCustomDataSetRec.Float (Const Fields :String) :Double;
  Begin
    Result := ghFloat (Value (Fields));
  End;

  Function TghCustomDataSetRec.Greater (Const Field1, Field2 :String)
    :Boolean;
  Begin
    Result := Value (Field1) > Value (Field2);
  End;

  Function TghCustomDataSetRec.HasField (Name :String;
    Const IncludeFieldObjs :Boolean = True;
    Const IncludeRefs :Boolean = True) :Boolean;
  Var
    Rec :TghCustomDataSetRec;
  Begin
    Name := CleanFieldName (Name);
    Result := InternalHasField (Name, IncludeFieldObjs) Or
      (IncludeRefs And GetFieldRec (Name, Rec, True) And (Rec <> Self) And
      Rec.HasField (Name, IncludeFieldObjs));
  End;

  Function TghCustomDataSetRec.HasFieldValue (Const Name :String)
    :Boolean;
  Begin
    Result := HasField (Name) And HasValue (Name);
  End;

  Function TghCustomDataSetRec.HasValue (Const Fields :String) :Boolean;
  Var
    AField :String;
    I :Integer;
    Rec :TghCustomDataSetRec;
  Begin
    I := 1;

    While GetNextField (Fields, I, Rec, AField) Do
      If (Rec = Nil) Or VarIsNull (Rec.GetValue (AField)) Then
      Begin
        Result := False;
        Exit;
      End;

    Result := True;
  End;

  Procedure TghCustomDataSetRec.Increment (Const AField :String;
    Const Inc :Double = 1; Const CheckNulls :Boolean = False);
  Begin
    If CheckNulls Then
      Values [AField] := Add (AField, Inc)
    Else
      Values [AField] := Float (AField) + Inc;
  End;

  Procedure TghCustomDataSetRec.Increment (Const AField, IncField :String;
    Const CheckNulls :Boolean = False);
  Begin
    If CheckNulls Then
      Values [AField] := Add (AField, IncField)
    Else
      Values [AField] := Float (AField) + Float (IncField);
  End;

  Function TghCustomDataSetRec.InRange (Const Value :Integer;
    Const MinField, MaxField :String) :Boolean;
  Begin
    Result := Math.InRange (Value, Int (MinField), Int (MaxField));
  End;

  Function TghCustomDataSetRec.InRange (
    Const Field, MinField, MaxField :String) :Boolean;
  Var
    AValue :Variant;
  Begin
    AValue := Value (Field);
    Result := (AValue >= Value (MinField)) And
      (AValue <= Value (MaxField));
  End;

  Function TghCustomDataSetRec.Int (Const Fields :String) :Integer;
  Begin
    Result := ghInt (Value (Fields));
  End;

  Function TghCustomDataSetRec.IntField (Const Name :String)
    :TIntegerField;
  Begin
    Result := Field (Name) As TIntegerField;
  End;

  Function TghCustomDataSetRec.IsGreater (Const Field1, Field2 :String)
    :Boolean;
  Begin
    Result := HasValue (Field1) And HasValue (Field2) And
      Greater (Field1, Field2);
  End;

  Function TghCustomDataSetRec.IsInRange (Const Value :Integer;
    Const MinField, MaxField :String) :Boolean;
  Begin
    Result := HasValue (MinField) And HasValue (MaxField) And
      InRange (Value, MinField, MaxField);
  End;

  Function TghCustomDataSetRec.IsInRange (
    Const Field, MinField, MaxField :String) :Boolean;
  Begin
    Result := HasValue (Field) And HasValue (MinField) And
      HasValue (MaxField) And InRange (Field, MinField, MaxField);
  End;

  Function TghCustomDataSetRec.IsLess (Const Field1, Field2 :String)
    :Boolean;
  Begin
    Result := HasValue (Field1) And HasValue (Field2) And
      Less (Field1, Field2);
  End;

  Function TghCustomDataSetRec.IsNull (Const Fields :String) :Boolean;
  Var
    AField :String;
    I :Integer;
    Rec :TghCustomDataSetRec;
  Begin
    I := 1;

    While GetNextField (Fields, I, Rec, AField) Do
      If (Rec <> Nil) And Not VarIsNull (Rec.GetValue (AField)) Then
      Begin
        Result := False;
        Exit;
      End;

    Result := True;
  End;

  Function TghCustomDataSetRec.Less (Const Field1, Field2 :String)
    :Boolean;
  Begin
    Result := Value (Field1) < Value (Field2);
  End;

  Function TghCustomDataSetRec.Max (Const Fields :String) :Variant;
  Var
    AField :String;
    AValue :Variant;
    I :Integer;
    Rec :TghCustomDataSetRec;
  Begin
    Result := Null;
    I := 1;

    While GetNextField (Fields, I, Rec, AField) Do
      If Rec <> Nil Then
      Begin
        AValue := Rec.GetValue (AField);

        If Not VarIsNull (AValue) And
        (VarIsNull (Result) Or (AValue > Result)) Then
          Result := AValue;
      End;
  End;

  Function TghCustomDataSetRec.Min (Const Fields :String) :Variant;
  Var
    AField :String;
    AValue :Variant;
    I :Integer;
    Rec :TghCustomDataSetRec;
  Begin
    Result := Null;
    I := 1;

    While GetNextField (Fields, I, Rec, AField) Do
      If Rec <> Nil Then
      Begin
        AValue := Rec.GetValue (AField);

        If Not VarIsNull (AValue) And
        (VarIsNull (Result) Or (AValue < Result)) Then
          Result := AValue;
      End;
  End;

  Function TghCustomDataSetRec.Multiply (Const AField :String;
    Const Factor :Double; Const Decimals :Integer = MaxInt) :Variant;
  Begin
    If IsNull (AField) Then
      Result := Null
    Else
      Result := ghRound (Value (AField) * Factor, Decimals);
  End;

  Function TghCustomDataSetRec.Multiply (
    Const AField, FactorField :String; Const Decimals :Integer = MaxInt)
    :Variant;
  Begin
    If IsNull (FactorField) Then
      Result := Null
    Else
      Result := Multiply (AField, Float (FactorField), Decimals);
  End;

  Function TghCustomDataSetRec.Percent (Const AField :String;
    Const Percentage :Double; Const Decimals :Integer = MaxInt) :Variant;
  Begin
    If IsNull (AField) Then
      Result := Null
    Else
      Result := ghPercent (Value (AField), Percentage, Decimals);
  End;

  Function TghCustomDataSetRec.Percent (
    Const AField, PercentageField :String;
    Const Decimals :Integer = MaxInt) :Variant;
  Begin
    If IsNull (PercentageField) Then
      Result := Null
    Else
      Result := Percent (AField, Float (PercentageField), Decimals);
  End;

  Function TghCustomDataSetRec.RefRec (AField :String)
    :TghCustomDataSetRec;
  Var
    Rec :TghCustomDataSetRec;
  Begin
    If GetFieldRec (AField, Rec) Then
      Result := Rec.InternalRefRec (AField)
    Else
      Result := Nil;
  End;

  Procedure TghCustomDataSetRec.Revert (Const Fields :String);
  Var
    AField :String;
    I :Integer;
    Rec :TghCustomDataSetRec;
  Begin
    I := 1;

    While GetNextField (Fields, I, Rec, AField) Do
      If Rec <> Nil Then
        Rec.SetValue (AField, Rec.GetValue ('?' + AField));
  End;

  Function TghCustomDataSetRec.Str (Const Fields :String) :String;
  Begin
    Result := VarToStr (Value (Fields));
  End;

  Function TghCustomDataSetRec.TwoInts (Const Field1, Field2 :String)
    :TghTwoInts;
  Begin
    Result := ghTwoInts (Int (Field1), Int (Field2));
  End;

  Function TghCustomDataSetRec.TwoInts (Const Fields :String)
    :TghTwoInts;
  Begin
    Result := ghTwoInts (Values [Fields]);
  End;

  ///revisar nombre
  Function TghCustomDataSetRec.Value (Const Fields :String) :Variant;
  Var
    AField :String;
    I :Integer;
    Rec :TghCustomDataSetRec;
  Begin
    I := 1;

    While GetNextField (Fields, I, Rec, AField) Do
      If Rec <> Nil Then
      Begin
        Result := Rec.GetValue (AField);

        If Not VarIsNull (Result) Then
          Exit;
      End;

    Result := Null;
  End;

  { Public virtual methods }

  Procedure TghCustomDataSetRec.Focus (Const AField :String);
  Var
    FieldObj :TField;
  Begin
    FieldObj := Field (AField);

    If FieldObj <> Nil Then
      FieldObj.FocusControl;
  End;

End.

