{*************************************************************************}
{ GH Freebrary                                                            }
{                                                                         }
{ GHFDataSetProvider.pas - TghDataSetProvider class unit.                 }
{                                                                         }
{ This software is distributed under the open source BSD license.         }
{                                                                         }
{ Copyright (c) 2002-2014, Al Gonzalez (rescatandoadelphi.blogspot.com)   }
{ All rights reserved.                                                    }
{                                                                         }
{ Redistribution and use in source and binary forms, with or without      }
{ modification, are permitted provided that the following conditions are  }
{ met:                                                                    }
{                                                                         }
{   - Redistributions of source code must retain the above copyright      }
{     notice, this list of conditions and the following disclaimer.       }
{   - Redistributions in binary form must reproduce the above copyright   }
{     notice, this list of conditions and the following disclaimer in the }
{     documentation and/or other materials provided with the              }
{     distribution.                                                       }
{   - Neither the name of Al Gonzalez nor the names of his contributors   }
{     may be used to endorse or promote products derived from this        }
{     software without specific prior written permission.                 }
{                                                                         }
{ THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS }
{ IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED   }
{ TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A         }
{ PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT      }
{ HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,  }
{ SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT        }
{ LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,   }
{ DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY   }
{ THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT     }
{ (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE   }
{ OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.    }
{*************************************************************************}

{ Please contribute to improve this library.  I am not a good programmer,
  but good programmers know what Delphi is worth.

  A.G. }

Unit GHFDataSetProvider;  { Data Set Provider }

Interface

  Uses
    Provider, Contnrs, Classes, DB, DBClient;

  Type
    { Data Set Provider class }
    TghDataSetProvider = Class (TDataSetProvider)
      Protected
        ClientID :String;
        Delegate :TghDataSetProvider;
        Delegates :TObjectList;
        FClientActivity :Boolean;
        FInnerDataSetClass :String;
        FMainTableName :String;
        GettingOperationParams :Boolean;
        LoadedDataSetActive :Boolean;

        { Static methods }
        Function CheckClientDelegated (Var OwnerData :OLEVariant) :Boolean;
        Procedure CheckFreeDelegate;
        Function ClientDelegate (Const ID :String) :TghDataSetProvider;
        Function Clone :TghDataSetProvider;
        Function DataSetStored :Boolean;
        Function IdleDelegate :TghDataSetProvider;
        Procedure ReadDataSetActive (Reader :TReader);
        Procedure SetClientActivity (Const Value :Boolean);
        Procedure SetDataSet (Const Value :TDataSet);
        Procedure SetInnerDataSetClass (Const Value :String);

        { Overridden methods }
        Procedure DefineProperties (Filer :TFiler); Override;
        Procedure DoAfterApplyUpdates (Var OwnerData :OLEVariant);
          Override;
        Procedure DoAfterExecute (Var OwnerData :OLEVariant); Override;
        Procedure DoAfterGetParams (Var OwnerData :OLEVariant); Override;
        Procedure DoAfterGetRecords (Var OwnerData :OLEVariant); Override;
        Procedure DoAfterRowRequest (Var OwnerData :OLEVariant); Override;
        Procedure DoBeforeApplyUpdates (Var OwnerData :OLEVariant);
          Override;
        Procedure DoBeforeExecute (Const CommandText :WideString;
          Var Params, OwnerData :OLEVariant); Override;
        Procedure DoBeforeGetParams (Var OwnerData :OLEVariant); Override;
        Procedure DoBeforeGetRecords (Count :Integer; Options :Integer;
          Const CommandText :WideString;
          Var Params, OwnerData :OLEVariant); Override;
        Procedure DoBeforeRowRequest (Var OwnerData :OLEVariant); Override;
        Procedure DoGetTableName (DataSet :TDataSet;
          Var ATableName :String); Override;
        Function GetChildOwner :TComponent; Override;
        Function GetChildParent :TComponent; Override;
        Procedure GetChildren (StoreRoutine :TGetChildProc;
          RootComponent :TComponent); Override;
        Function InternalApplyUpdates (Const Delta :OLEVariant;
          MaxErrors :Integer; Out ErrorCount :Integer) :OLEVariant;
          Override;
        Procedure InternalExecute (Const CommandText :WideString;
          Var Params :OLEVariant); Override;
        Function InternalGetParams (Types :TParamTypes = AllParamTypes)
          :OLEVariant; Override;
        Function InternalGetRecords (Count :Integer; Out RecsOut :Integer;
          Options :TGetRecordOptions; Const CommandText :WideString;
          Var Params :OLEVariant) :OLEVariant; Override;
        Function InternalRowRequest (Const Row :OLEVariant;
          RequestType :TFetchOptions): OLEVariant; Override;
        Procedure Loaded; Override;
        Procedure SetChildOrder (Component :TComponent; Order :Integer);
          Override;
      Public
        Destructor Destroy; Override;

        { Static methods }
        Function GetTableName :String;

        { Overridden methods }
        Function DataRequest (Input :OLEVariant) :OLEVariant; Override;
      Published
        { Properties }
        Property ClientActivity :Boolean Read FClientActivity
          Write SetClientActivity Default False;
        Property DataSet Write SetDataSet Stored DataSetStored;
        Property InnerDataSetClass :String Read FInnerDataSetClass
          Write SetInnerDataSetClass;
        Property MainTableName :String Read FMainTableName Write FMainTableName;
    End;

Implementation

  Uses
    SysUtils, GHFRTL, GHFVCL, TypInfo;

  { TghDataSetProvider }

  Destructor TghDataSetProvider.Destroy;
  Begin
    FreeAndNil (Delegates);
    Inherited Destroy;
  End;

  { Protected static methods }

  {$Warn No_RetVal Off}
  Function TghDataSetProvider.CheckClientDelegated (
    Var OwnerData :OLEVariant) :Boolean;
  Begin
    If ClientActivity Then
      Try
        If (ClientID <> '') And (ClientID <> OwnerData [0]) Then
        Begin
          Result := True;

          { Search corresponding delegate provider }
          Delegate := ClientDelegate (OwnerData [0]);

          If Delegate <> Nil Then
            Exit;

          If DataSet.Active Then
          Begin
            { Search first idle delegate provider }
            Delegate := IdleDelegate ();

            If Delegate <> Nil Then
              Exit;

            { Create a delegate provider }
            {****************************}
            Delegate := Clone ();

            Try
              Delegate.ClientActivity := False;
              Delegate.ClientID := OwnerData [0];

              If Delegates = Nil Then
                Delegates := TObjectList.Create;

              Delegates.Add (Delegate);
            Except
              CheckFreeDelegate;
              Raise;
            End;
            {****************************}

            Exit;
          End;
        End;

        ClientID := OwnerData [0];
      Finally
        OwnerData := OwnerData [1];
      End;

    Delegate := Nil;
    Result := False;
  End;
  {$Warn No_RetVal On}

  Procedure TghDataSetProvider.CheckFreeDelegate;
  Begin
    If (Delegate <> Nil) And Not Delegate.DataSet.Active Then
      If (Delegates <> Nil) And (Delegates.Remove (Delegate) > -1) Then
        Delegate := Nil
      Else
        FreeAndNil (Delegate);
  End;

  Function TghDataSetProvider.ClientDelegate (Const ID :String)
    :TghDataSetProvider;
  Var
    I :Integer;
  Begin
    If Delegates <> Nil Then
      For I := 0 To Delegates.Count - 1 Do
        If TghDataSetProvider (Delegates [I]).ClientID = ID Then
        Begin
          Result := TghDataSetProvider (Delegates [I]);
          Exit;
        End;

    Result := Nil;
  End;

  Function TghDataSetProvider.Clone :TghDataSetProvider;
  Begin
    Result := TghDataSetProvider (ghClone (Self, Self, 'DataSet;Name'));

    Try
      If InnerDataSetClass = '' Then
        Result.DataSet := TDataSet (ghCloneDataSet (
          DataSet, Result, 'Active'))
      Else
        ghCopyDataSetProps (DataSet, Result.DataSet, 'Active');
    Except
      Result.Free;
      Raise;
    End;
  End;

  Function TghDataSetProvider.DataSetStored :Boolean;
  Begin
    { The DataSet property is stored in a special way when that object is
      held internally (InnerDataSetClass <> '').  See also the
      DefineProperties method. }
    Result := InnerDataSetClass = '';
  End;

  Function TghDataSetProvider.IdleDelegate :TghDataSetProvider;
  Var
    I :Integer;
  Begin
    If Delegates <> Nil Then
      For I := 0 To Delegates.Count - 1 Do
        With TghDataSetProvider (Delegates [I]) Do
          If (DataSet <> Nil) And Not DataSet.Active Then
          Begin
            Result := TghDataSetProvider (Delegates [I]);
            Exit;
          End;

    Result := Nil;
  End;

  Procedure TghDataSetProvider.ReadDataSetActive (Reader :TReader);
  Begin
    LoadedDataSetActive := Reader.ReadBoolean;
  End;

  Procedure TghDataSetProvider.SetClientActivity (Const Value :Boolean);
  Begin
    FClientActivity := Value;

    If Not Value Then
      FreeAndNil (Delegates);
  End;

  Procedure TghDataSetProvider.SetDataSet (Const Value :TDataSet);
  Begin
    If Value = DataSet Then
      Exit;

    InnerDataSetClass := '';
    Inherited DataSet := Value;

    If (ComponentState * [csLoading, csDesigning] = [csDesigning]) And
    (DataSet <> Nil) And (MainTableName = '') Then
      MainTableName := ghTable (DataSet);
  End;

  Procedure TghDataSetProvider.SetInnerDataSetClass (
    Const Value :String);
  Begin
    If Value = InnerDataSetClass Then
      Exit;

    If InnerDataSetClass <> '' Then
    Begin
      DataSet.Free;  // The Notification method sets DataSet to Nil
      FInnerDataSetClass := '';
    End;

    If (Value <> '') And FindClass (Value).InheritsFrom (TDataSet) Then
    Begin
      Inherited DataSet := TDataSetClass (GetClass (Value)).Create (Self);
      ghSetSubcomponent (DataSet, 'DataSet');
      FInnerDataSetClass := Value;
    End;
  End;

  { Protected overridden methods }

  { The DataSet property is stored in a special way when that object is
    held internally (InnerDataSetClass <> '').  In this case, the
    overridden DefineProperties method allows the DataSet properties to be
    stored after the InnerDataSetClass property.  Thus, when the component
    is recreated, the DataSet instance will exist before its properties are
    read. }
  Type
    TDataSetAccess = Class (TDataSet);
    TWriterAccess = Class (TWriter);
  Procedure TghDataSetProvider.DefineProperties (Filer :TFiler);
  Var
    DataSetState :TDataSetState;
    WriteDataSetActive :Boolean;
  Begin
    Inherited DefineProperties (Filer);

    If Filer Is TReader Then
      Filer.DefineProperty ('DataSetActive', ReadDataSetActive, Nil, False)
    Else
      If (Filer Is TWriter) And (InnerDataSetClass <> '') Then
      Begin
        If (Filer.Ancestor = Nil) Or
        (TghDataSetProvider (Filer.Ancestor).DataSet = Nil) Then
          DataSetState := dsInactive
        Else
          DataSetState := TghDataSetProvider (
            Filer.Ancestor).DataSet.State;

        WriteDataSetActive := (DataSet.State <> DataSetState) And
          (Not IsPublishedProp (DataSet, 'StoreActive') Or
          ghPropBool (DataSet, 'StoreActive'));

        If WriteDataSetActive Then
          { We avoid store the DataSet.Active property.  NOTE: When
            TReader.ReadComponent calls
            "SetSubComponentState([csReading])", the DataSet property is
            still Nil; the csReading state is necessary for the
            TDataSet.SetActive method at design time. }
          DataSetState := TDataSetAccess (DataSet).SetTempState (
            DataSetState);

        Try
          TWriterAccess (Filer).WriteProperty (Self,
            GetPropInfo (Self, 'DataSet'));
        Finally
          If WriteDataSetActive Then
            TDataSetAccess (DataSet).RestoreState (DataSetState);
        End;

        If WriteDataSetActive Then
        Begin
          // DataSetActive replaces the DataSet.Active property
          TWriterAccess (Filer).WritePropName ('DataSetActive');
          TWriter (Filer).WriteBoolean (DataSet.Active);
        End;
      End;
  End;

  Procedure TghDataSetProvider.DoAfterApplyUpdates (
    Var OwnerData :OLEVariant);
  Begin
    If Delegate <> Nil Then
      Try
        Delegate.DoAfterApplyUpdates (OwnerData);
      Finally
        CheckFreeDelegate;
      End
    Else
      Inherited DoAfterApplyUpdates (OwnerData);
  End;

  Procedure TghDataSetProvider.DoAfterExecute (
    Var OwnerData :OLEVariant);
  Begin
    If Delegate <> Nil Then
      Try
        Delegate.DoAfterExecute (OwnerData);
      Except
        CheckFreeDelegate;
        Raise;
      End
    Else
      Inherited DoAfterExecute (OwnerData);

    GettingOperationParams := True;
  End;

  Procedure TghDataSetProvider.DoAfterGetParams (
    Var OwnerData :OLEVariant);
  Begin
    If Delegate <> Nil Then
      Try
        Delegate.DoAfterGetParams (OwnerData);
      Finally
        CheckFreeDelegate;
      End
    Else
      Inherited DoAfterGetParams (OwnerData);
  End;

  Procedure TghDataSetProvider.DoAfterGetRecords (
    Var OwnerData :OLEVariant);
  Begin
    If Delegate <> Nil Then
      Try
        Delegate.DoAfterGetRecords (OwnerData);
      Except
        CheckFreeDelegate;
        Raise;
      End
    Else
      Inherited DoAfterGetRecords (OwnerData);

    GettingOperationParams := True;
  End;

  Procedure TghDataSetProvider.DoAfterRowRequest (
    Var OwnerData :OLEVariant);
  Begin
    If Delegate <> Nil Then
      Try
        Delegate.DoAfterRowRequest (OwnerData);
      Finally
        CheckFreeDelegate;
      End
    Else
      Inherited DoAfterRowRequest (OwnerData);
  End;

  Procedure TghDataSetProvider.DoBeforeApplyUpdates (
    Var OwnerData :OLEVariant);
  Begin
    If CheckClientDelegated (OwnerData) Then
      Try
        Delegate.DoBeforeApplyUpdates (OwnerData);
      Except
        CheckFreeDelegate;
        Raise;
      End
    Else
      Inherited DoBeforeApplyUpdates (OwnerData);
  End;

  Procedure TghDataSetProvider.DoBeforeExecute (
    Const CommandText :WideString; Var Params, OwnerData :OLEVariant);
  Begin
    If CheckClientDelegated (OwnerData) Then
      Try
        Delegate.DoBeforeExecute (CommandText, Params, OwnerData);
      Except
        CheckFreeDelegate;
        Raise;
      End
    Else
      Inherited DoBeforeExecute (CommandText, Params, OwnerData);
  End;

  Procedure TghDataSetProvider.DoBeforeGetParams (
    Var OwnerData :OLEVariant);
  Begin
    If CheckClientDelegated (OwnerData) Then
      Try
        Delegate.DoBeforeGetParams (OwnerData);
      Except
        CheckFreeDelegate;
        Raise;
      End
    Else
      Inherited DoBeforeGetParams (OwnerData);

    GettingOperationParams := False;
  End;

  Procedure TghDataSetProvider.DoBeforeGetRecords (Count :Integer;
    Options :Integer; Const CommandText :WideString;
    Var Params, OwnerData :OLEVariant);
  Begin
    If CheckClientDelegated (OwnerData) Then
      Try
        Delegate.DoBeforeGetRecords (
          Count, Options, CommandText, Params, OwnerData);
      Except
        CheckFreeDelegate;
        Raise;
      End
    Else
      Inherited DoBeforeGetRecords (
        Count, Options, CommandText, Params, OwnerData);
  End;

  Procedure TghDataSetProvider.DoBeforeRowRequest (
    Var OwnerData :OLEVariant);
  Begin
    If CheckClientDelegated (OwnerData) Then
      Try
        Delegate.DoBeforeRowRequest (OwnerData);
      Except
        CheckFreeDelegate;
        Raise;
      End
    Else
      Inherited DoBeforeRowRequest (OwnerData);
  End;

  Procedure TghDataSetProvider.DoGetTableName (DataSet :TDataSet;
    Var ATableName :String);
  Begin
    {}//probar
    If (DataSet = Self.DataSet) And (MainTableName <> '') Then
      ATableName := MainTableName;

    Inherited DoGetTableName (DataSet, ATableName);
  End;

  Function TghDataSetProvider.GetChildOwner :TComponent;
  Begin
    { NOTE: The TFieldsEditor class shows the field objets list only if
      they share the same owner as the data set.  Due to the
      TghDataSetProvider component and not the root component is the
      owner of the field objets when using an inner data set
      (InnerDataSetClass <> ''), it is not possible modify, at design time,
      field objects inherited from another data modules (the Delphi 7
      designer does not support reading of inherited objects if they do not
      directly belong to the root component). }
    Result := Self;
  End;

  Function TghDataSetProvider.GetChildParent :TComponent;
  Begin
    Result := DataSet;
  End;

  Procedure TghDataSetProvider.GetChildren (
    StoreRoutine :TGetChildProc; RootComponent :TComponent);
  Var
    I :Integer;
  Begin
    If InnerDataSetClass <> '' Then
      With DataSet Do
      Begin
        For I := 0 To Fields.Count - 1 Do
          If Fields [I].Owner = Self Then
            StoreRoutine (Fields [I]);

        For I := 0 To AggFields.Count - 1 Do
          If AggFields [I].Owner = Self Then
            StoreRoutine (AggFields [I]);
      End;
  End;

  Function TghDataSetProvider.InternalApplyUpdates (
    Const Delta :OLEVariant; MaxErrors :Integer; Out ErrorCount :Integer)
    :OLEVariant;
  Begin
    If Delegate <> Nil Then
      Try
        Result := Delegate.InternalApplyUpdates (
          Delta, MaxErrors, ErrorCount);
      Except
        CheckFreeDelegate;
        Raise;
      End
    Else
      Result := Inherited InternalApplyUpdates (
        Delta, MaxErrors, ErrorCount)
  End;

  Procedure TghDataSetProvider.InternalExecute (
    Const CommandText :WideString; Var Params :OLEVariant);
  Begin
    If Delegate <> Nil Then
      Try
        Delegate.InternalExecute (CommandText, Params);
      Except
        CheckFreeDelegate;
        Raise;
      End
    Else
      Inherited InternalExecute (CommandText, Params);
  End;

  Function TghDataSetProvider.InternalGetParams (
    Types :TParamTypes = AllParamTypes) :OLEVariant;
  Begin
    If Delegate <> Nil Then
      Try
        Try
          Result := Delegate.InternalGetParams (Types);
        Except
          CheckFreeDelegate;
          Raise;
        End
      Finally
        If GettingOperationParams Then
          CheckFreeDelegate;
      End
    Else
      Result := Inherited InternalGetParams (Types);
  End;

  Function TghDataSetProvider.InternalGetRecords (Count :Integer;
    Out RecsOut :Integer; Options :TGetRecordOptions;
    Const CommandText :WideString; Var Params :OLEVariant) :OLEVariant;
  Begin
    If Delegate <> Nil Then
      Try
        Result := Delegate.InternalGetRecords (
          Count, RecsOut, Options, CommandText, Params)
      Except
        CheckFreeDelegate;
        Raise;
      End
    Else
      Result := Inherited InternalGetRecords (
        Count, RecsOut, Options, CommandText, Params);
  End;

  Function TghDataSetProvider.InternalRowRequest (Const Row :OLEVariant;
    RequestType :TFetchOptions): OLEVariant;
  Begin
    If Delegate <> Nil Then
      Try
        Result := Delegate.InternalRowRequest (Row, RequestType);
      Except
        CheckFreeDelegate;
        Raise;
      End
    Else
      Result := Inherited InternalRowRequest (Row, RequestType);
  End;

  Procedure TghDataSetProvider.Loaded;
  Begin
    Inherited Loaded;

    { If the data set is inner, we assign to it its stored state (
      DataSetActive seudo-property) }
    If InnerDataSetClass <> '' Then
      DataSet.Active := LoadedDataSetActive;
  End;

  Procedure TghDataSetProvider.SetChildOrder (Component :TComponent;
    Order :Integer);
  Begin
    If (InnerDataSetClass <> '') And (Component Is TField) And
    (DataSet.Fields.IndexOf (TField (Component)) > -1) Then
      TField (Component).Index := Order;
  End;

  { Public static methods }
  
  Function TghDataSetProvider.GetTableName :String;
  Begin
    DoGetTableName (Dataset, Result);
  End;

  { Public overridden methods }

  Function TghDataSetProvider.DataRequest (Input :OLEVariant)
    :OLEVariant;
  Begin
    If CheckClientDelegated (Input) Then
      Try
        Result := Delegate.DataRequest (Input);
      Finally
        CheckFreeDelegate;
      End
    Else
      Result := Inherited DataRequest (Input);
  End;

End.

