{*************************************************************************}
{ GH Freebrary                                                            }
{                                                                         }
{ GHFEN.pas - Unit of elements related to the English language.           }
{                                                                         }
{ This software is distributed under the open source BSD license.         }
{                                                                         }
{ Copyright (c) 2002-2014, Al Gonzalez (rescatandoadelphi.blogspot.com)   }
{ All rights reserved.                                                    }
{                                                                         }
{ Redistribution and use in source and binary forms, with or without      }
{ modification, are permitted provided that the following conditions are  }
{ met:                                                                    }
{                                                                         }
{   - Redistributions of source code must retain the above copyright      }
{     notice, this list of conditions and the following disclaimer.       }
{   - Redistributions in binary form must reproduce the above copyright   }
{     notice, this list of conditions and the following disclaimer in the }
{     documentation and/or other materials provided with the              }
{     distribution.                                                       }
{   - Neither the name of Al Gonzalez nor the names of his contributors   }
{     may be used to endorse or promote products derived from this        }
{     software without specific prior written permission.                 }
{                                                                         }
{ THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS }
{ IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED   }
{ TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A         }
{ PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT      }
{ HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,  }
{ SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT        }
{ LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,   }
{ DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY   }
{ THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT     }
{ (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE   }
{ OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.    }
{*************************************************************************}

{ Please contribute to improve this library.  I am not a good programmer,
  but good programmers know what Delphi is worth.

  A.G. }

Unit GHFEN;  { English }

Interface

  Const
    { Letters }

    ghltENLowers = ['a'..'z'];  // Lower case
    ghltENUppers = ['A'..'Z'];  // Upper case
    ghltENAll    = ghltENLowers + ghltENUppers;

    // Vowels
    ghltENVowelsLower = ['a', 'e', 'i', 'o', 'u'];  // Lower case
    ghltENVowelsUpper = ['A', 'E', 'I', 'O', 'U'];  // Upper case
    ghltENVowels      = ghltENVowelsLower + ghltENVowelsUpper;

    // Consonants

    // Lower case
    ghltENConsonantsLower = ghltENLowers - ghltENVowelsLower;

    // Upper case
    ghltENConsonantsUpper = ghltENUppers - ghltENVowelsUpper;

    ghltENConsonants      = ghltENConsonantsLower + ghltENConsonantsUpper;

  { Functions and procedures }

  { English Lower case }
  Function ghENLower (Const Value :Char) :Char;

  { English Set Lower case }
  Function ghENSetLower (Var Str :String; Const Index :Integer = 1)
    :Boolean; Overload;

  { English Set Lower case }
  Procedure ghENSetLower (Var Strs :Array Of String;
    Const Index :Integer = 1); Overload;

  { English Set Lower case Safe }
  Function ghENSetLowerSafe (Var Str :String; Const Index :Integer = 1)
    :Boolean; Overload;

  { English Set Lower case Safe }
  Procedure ghENSetLowerSafe (Var Strs :Array Of String;
    Const Index :Integer = 1); Overload;

Implementation

  Uses
    GHFRTL;

  { Functions and procedures }

  // Based on the System.UpCase function
  Function ghENLower (Const Value :Char) :Char;
  Asm
    Cmp AL, 'A'
    JB  @Exit
    Cmp AL, 'Z'
    JA  @Exit
    Add AL, 'a' - 'A'
    @Exit:
  End;

  Function ghENSetLower (Var Str :String; Const Index :Integer = 1)
    :Boolean; Overload;
  Begin
    Result := ghSetChr (Str, Index, ghENLower (Str [Index]));
  End;

  Procedure ghENSetLower (Var Strs :Array Of String;
    Const Index :Integer = 1); Overload;
  Var
    I :Integer;
  Begin
    For I := 0 To High (Strs) Do
      ghENSetLower (Strs [I], Index);
  End;

  Function ghENSetLowerSafe (Var Str :String; Const Index :Integer = 1)
    :Boolean; Overload;
  Begin
    Result := ghIsIndex (Index, Str) And ghENSetLower (Str, Index);
  End;

  Procedure ghENSetLowerSafe (Var Strs :Array Of String;
    Const Index :Integer = 1); Overload;
  Var
    I :Integer;
  Begin
    For I := 0 To High (Strs) Do
      ghENSetLowerSafe (Strs [I], Index);
  End;

End.

