{*************************************************************************}
{ GH Freebrary                                                            }
{                                                                         }
{ GHFES.pas - Unit of elements related to the Spanish language.           }
{                                                                         }
{ This software is distributed under the open source BSD license.         }
{                                                                         }
{ Copyright (c) 2002-2014, Al Gonzalez (rescatandoadelphi.blogspot.com)   }
{ All rights reserved.                                                    }
{                                                                         }
{ Redistribution and use in source and binary forms, with or without      }
{ modification, are permitted provided that the following conditions are  }
{ met:                                                                    }
{                                                                         }
{   - Redistributions of source code must retain the above copyright      }
{     notice, this list of conditions and the following disclaimer.       }
{   - Redistributions in binary form must reproduce the above copyright   }
{     notice, this list of conditions and the following disclaimer in the }
{     documentation and/or other materials provided with the              }
{     distribution.                                                       }
{   - Neither the name of Al Gonzalez nor the names of his contributors   }
{     may be used to endorse or promote products derived from this        }
{     software without specific prior written permission.                 }
{                                                                         }
{ THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS }
{ IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED   }
{ TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A         }
{ PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT      }
{ HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,  }
{ SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT        }
{ LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,   }
{ DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY   }
{ THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT     }
{ (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE   }
{ OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.    }
{*************************************************************************}

{ Please contribute to improve this library.  I am not a good programmer,
  but good programmers know what Delphi is worth.

  A.G. }

Unit GHFES;  { Spanish }

Interface

  Uses
    GHFEN, GHFRTL;

  Const
    { Letters }

    // Accent
    ghltESAccentsLower = ['', '', '', '', ''];  // Lower case
    ghltESAccentsUpper = ['', '', '', '', ''];  // Upper case
    ghltESAccents      = ghltESAccentsLower + ghltESAccentsUpper;

    // Simple
    ghltESSimplesLower = ghltENLowers + [''];  // Lower case
    ghltESSimplesUpper = ghltENUppers + [''];  // Upper case
    ghltESSimples      = ghltESSimplesLower + ghltESSimplesUpper;

    // Lower case
    ghltESLowers = ghltESSimplesLower + ghltESAccentsLower + [''];

    // Upper case
    ghltESUppers = ghltESSimplesUpper + ghltESAccentsUpper + [''];

    ghltESAll    = ghltESLowers + ghltESUppers;

    // Vowels

    ghltESAs = ['a', 'A', '', ''];
    ghltESEs = ['e', 'E', '', ''];
    ghltESIs = ['i', 'I', '', ''];
    ghltESOs = ['o', 'o', '', ''];
    ghltESUs = ['u', 'U', '', ''];

    // Simple
    ghltESSimpleVowelsLower = ghltENVowelsLower;  // Lower case
    ghltESSimpleVowelsUpper = ghltENVowelsUpper;  // Upper case
    ghltESSimpleVowels      = ghltESSimpleVowelsLower +
                              ghltESSimpleVowelsUpper;

    // Lower case
    ghltESVowelsLower = ghltESSimpleVowelsLower + ghltESAccentsLower +
                        [''];

    // Upper case
    ghltESVowelsUpper = ghltESSimpleVowelsUpper + ghltESAccentsUpper +
                        [''];

    ghltESVowels      = ghltESVowelsLower + ghltESVowelsUpper;

    // Open (a, e, o)

    // Accent
    ghltESAccentOpenVowelsLower = ['', '', ''];  // Lower case
    ghltESAccentOpenVowelsUpper = ['', '', ''];   // Upper case
    ghltESAccentOpenVowels      = ghltESAccentOpenVowelsLower +
                                  ghltESAccentOpenVowelsUpper;

    // Simple
    ghltESSimpleOpenVowelsLower = ['a', 'e', 'o'];  // Lower case
    ghltESSimpleOpenVowelsUpper = ['A', 'E', 'O'];  // Upper case
    ghltESSimpleOpenVowels      = ghltESSimpleOpenVowelsLower +
                                  ghltESSimpleOpenVowelsUpper;

    ghltESOpenVowelsLower = ghltESSimpleOpenVowelsLower +
                            ghltESAccentOpenVowelsLower;  // Lower case
    ghltESOpenVowelsUpper = ghltESSimpleOpenVowelsUpper +
                            ghltESAccentOpenVowelsUpper;  // Upper case
    ghltESOpenVowels      = ghltESOpenVowelsLower + ghltESOpenVowelsUpper;

    // Closed (i, u)

    // Accent
    ghltESAccentClosedVowelsLower = ['', ''];  // Lower case
    ghltESAccentClosedVowelsUpper = ['', ''];  // Upper case
    ghltESAccentClosedVowels      = ghltESAccentClosedVowelsLower +
                                    ghltESAccentClosedVowelsUpper;

    // Simple
    ghltESSimpleClosedVowelsLower = ['i', 'u'];  // Lower case
    ghltESSimpleClosedVowelsUpper = ['I', 'U'];  // Upper case
    ghltESSimpleClosedVowels      = ghltESSimpleClosedVowelsLower +
                                    ghltESSimpleClosedVowelsUpper;

    // Lower case
    ghltESClosedVowelsLower = ghltESSimpleClosedVowelsLower +
                              ghltESAccentClosedVowelsLower + [''];

    // Upper case
    ghltESClosedVowelsUpper = ghltESSimpleClosedVowelsUpper +
                              ghltESAccentClosedVowelsUpper + [''];

    ghltESClosedVowels      = ghltESClosedVowelsLower +
                              ghltESClosedVowelsUpper;

    // Consonants

    // Lower case
    ghltESConsonantsLower = ghltESLowers - ghltESVowelsLower;

    // Upper case
    ghltESConsonantsUpper = ghltESUppers - ghltESVowelsUpper;
    
    ghltESConsonants      = ghltESConsonantsLower + ghltESConsonantsUpper;

  Var
    { Spanish Singular ("virtual" function) }
    GHESSingular :TghStrParamStrFunc;

    { Spanish Singular Word ("virtual" function) }
    GHESSingularWord :TghStrParamStrFunc;

    { Spanish Syllables ("virtual" function) }
    GHESSyllables :TghStrParamIntFunc;

  { Functions and procedures }

  { Spanish A ("un") }
  Function ghESA (Const Gender :TghGender; Const Plural :Boolean = False)
    :String;

  { Spanish Ending }
  Function ghESEnding (Const Gender :TghGender;
    Const Plural :Boolean = False) :String;

  { Spanish Ending and That }
  Function ghESEndingThat (Const Gender :TghGender;
    Const Plural :Boolean = False) :String;

  { Spanish In Letters }
  Function ghESInLetters (Value :Int64) :String;

  { Spanish In Letters Quantity }
  Function ghESInLettersQuantity (Const Value :Int64;
    Const Gender :TghGender = ghgnMasculine) :String;

  { Spanish Is Diphthong? }
  Function ghESIsDiphthong (Const Letter1, Letter2 :Char) :Boolean;

  { Spanish That ("ese") }
  Function ghESThat (Const Gender :TghGender;
    Const Plural :Boolean = False) :String;

  { Spanish The ("el") }
  Function ghESThe (Const Gender :TghGender; Const Plural :Boolean = False)
    :String;

  { Spanish To The ("al") }
  Function ghESToThe (Const Gender :TghGender;
    Const Plural :Boolean = False) :String;

Implementation

  Uses
    SysUtils, Windows, StrUtils;

  { Functions and procedures }

  Function BasicSyllables (Const Value :String) :Integer;
  Var
    I :Integer;
    PrevChr :Char;
  Begin
    PrevChr := #0;
    Result := 0;

    For I := 1 To Length (Value) Do
    Begin
      If (Value [I] In ghltESVowels) And
      Not ghESIsDiphthong (PrevChr, Value [I]) Then
        Inc (Result);

      If Not (Value [I] In ['h', 'H']) Then
        PrevChr := Value [I];
    End;
  End;

  Function BasicSingular (Const Value :String) :String;
  Begin
    Result := ghSplitLast (Value);
    Result := ghDeleteRight (Value, Length (Result)) +
      GHESSingularWord (Result);
  End;

  Function BasicSingularWord (Const Value :String) :String;
  Const
    CommonVowels = ghltESSimpleVowels + ['', '', '', ''];
  Var
    AccentIndex :Integer;

    Function CheckInvariable (Const Words :Array Of String) :Boolean;
    Var
      I :Integer;
    Begin
      For I := 0 To High (Words) Do
        If ghEquals (Value, Words [I]) Then
        Begin
          BasicSingularWord := Value;
          Result := True;
          Exit;
        End;

      Result := False;
    End;

    Function CheckSpecial (Const Words :Array Of String) :Boolean;
    Var
      I :Integer;
    Begin
      For I := 0 To High (Words) Div 2 Do
        If ghEquals (Value, Words [I * 2]) Then
        Begin
          BasicSingularWord := Words [(I * 2) + 1];
          Result := True;
          Exit;
        End;

      Result := False;
    End;

    Function CheckTruncate (Const Words :Array Of String;
      Const Chrs :Integer = 1) :Boolean;
    Var
      I :Integer;
    Begin
      For I := 0 To High (Words) Do
        If ghEquals (Value, Words [I]) Then
        Begin
          BasicSingularWord := ghDeleteRight (Value, Chrs);
          Result := True;
          Exit;
        End;

      Result := False;
    End;

    Function Ends (Const Chrs :TSysCharSet; Const SubStr :String) :Boolean;
    Begin
      If Chrs <> [] Then
        Result := ghEnds (Value, Chrs, SubStr, ghscOrdAnyCase)
      Else
        Result := ghEnds (Value, SubStr, ghscOrdAnyCase);
    End;
  Begin
    { NOTE: This function assumes that the Value parameter is a noun or
      adjective in plural form.  It also supports some articles and
      pronouns. } 

    // ...Vowel a/e/i//o/u/ + nes -> ...Vowel + n
    If Ends (CommonVowels, 'nes') Then
    Begin
      // Exceptions (special plurals)
      If Not CheckSpecial (['especmenes', 'espcimen', 'regmenes',
      'rgimen']) And

      // Exceptions (invariable form)
      Not CheckInvariable (['alcatenes', 'cascapiones', 'demstenes',
      'entrepanes', 'entrepeines', 'folijones', 'grasones', 'guadafiones',
      'hipmanes', 'lunes', 'maitines', 'manes', 'portaaviones',
      'quijones', 'reviernes', 'viernes']) And

      // Exceptions (...vowel + ne)
      Not CheckTruncate (['lsines', 'amaines', 'argemones', 'calgines',
      'cicerones', 'cines', 'desgranes', 'dmines', 'empeines', 'enes',
      'epmones', 'higienes', 'impunes', 'inanes', 'inmunes', 'nenes',
      'paraselenes', 'peines', 'penes', 'perennes', 'piunes',
      'sobreempeines', 'tiznes', 'vorgines']) Then
      Begin
        Result := ghDeleteRight (Value, 2);  // panes -> pan
        AccentIndex := ghIndex (Result, ghltESAccents);

        If AccentIndex > 0 Then
        Begin
          If Not ANSISameText (Result, 'san') Then
            // Singular without accent (imgenes -> imagen)
            ghSimplify (Result, AccentIndex);
        End
        Else
          If GHESSyllables (Result) > 1 Then
            // Singular with accent (camiones -> camin)
            ghAccentEnd (Result, 2);
      End;
    End
    Else
      // ...Vowel a/e/i//o/u/ + res -> ...Vowel + r
      If Ends (CommonVowels, 'res') Then
      Begin
        // Exceptions (special plurals)
        If Not CheckSpecial (['caracteres', 'carcter', 'lores', 'lord',
        'milores', 'milord']) And

        // Exceptions (invariable form)
        Not CheckInvariable (['achares', 'acores', 'chiqueadores',
        'enseres', 'milamores', 'pesalicores', 'traeres', 'vveres']) And

        // Exceptions (...vowel + re)
        Not CheckTruncate (['acures', 'aires', 'alboaires', 'apures',
        'ataires', 'bucares', 'bcares', 'cleres', 'congneres',
        'curares', 'desaires', 'desgaires', 'donaires', 'eres',
        'folclores', 'manares', 'mapanares', 'maures', 'mavacures',
        'misereres', 'naires', 'pelaires', 'rabasaires', 'socaires',
        'tentenelaires', 'tteres', 'trabucaires', 'yares']) Then
          Result := ghDeleteRight (Value, 2);  // crteres -> crter
      End
      Else
        // ...Vowel a/e/i//o/u/ + les -> ...Vowel + l
        If Ends (CommonVowels, 'les') Then
        Begin
          // Exceptions (invariable form)
          If Not CheckInvariable (['anofeles', 'esponsales', 'grafioles',
          'guardamateriales', 'gules', 'hrcules', 'lupercales',
          'mircoles', 'moriles', 'onoquiles', 'pisapapeles',
          'portapapeles', 'salvamanteles', 'sujetapapeles']) And

          // Exceptions (...vowel + le)
          Not CheckTruncate (['acaules', 'amoles', 'atoles', 'bailes',
          'balele', 'chiles', 'deles', 'desfiles', 'distoles', 'eles',
          'extrasstoles', 'frailes', 'freile', 'hidroceles', 'hiprboles',
          'hules', 'imbeles', 'ndoles', 'moles', 'multicaules',
          'narguiles', 'paradistoles', 'peleles', 'perstoles', 'pinoles',
          'pozoles', 'proles', 'sanciroles', 'sarcoceles', 'sstoles',
          'toles', 'triles', 'tules', 'unicaule']) Then
            Result := ghDeleteRight (Value, 2);  // esenciales -> esencial
        End
        Else
          // ...Vowel a/e//u/ + des -> ...Vowel + d
          If Ends (['a', 'A', 'e', 'E', '', '', 'u', 'U', '', ''],
          'des') Then
          Begin
            // Exception (invariable form)
            If SameText (Value, 'tientaparedes') Then
              Result := Value
            Else
              // Exceptions (...vowel + de)
              If Not CheckTruncate (['almohades', 'rcades',
              'archicofrades', 'cofrades', 'fraudes', 'jades', 'laudes',
              'leudes', 'mnades', 'nyades', 'parstades', 'plyades',
              'sedes', 'trbedes']) Then
                Result := ghDeleteRight (Value, 2);  // virtudes -> virtud
          End
          Else
            // ...Vowel a/e/i//o/u + ces -> ...Vowel + z
            If Ends (ghltESSimpleVowels + ['', ''], 'ces') Then
            Begin
              // Exceptions (invariable form)
              If Not CheckInvariable (['actinomices', 'bruces',
              'cascanueces', 'cortalpices', 'creces', 'fasces', 'fauces',
              'grmeces', 'preces']) And

              // Exceptions (...vowel + ce)
              Not CheckTruncate (['aleces', 'apndices', 'pices',
              'artfices', 'arspices', 'cauces', 'cdices', 'cmplices',
              'desenlaces', 'desguaces', 'desguinces', 'doces', 'dplices',
              'engaces', 'enlaces', 'esconces', 'esguinces', 'espdices',
              'fices', 'frunces', 'frtices', 'goces', 'hlices', 'bices',
              'ndices', 'lances', 'lrices', 'multplices', 'mrices',
              'bices', 'nices', 'orfices', 'pnaces', 'plices',
              'pontfice', 'roces', 'sardnices', 'sauces', 'slices',
              'subndices', 'treces', 'trplices', 'vrtices',
              'vrtices']) Then
                Result := ghDeleteRight (Value, 3) +
                  ghCased ('z', ghCase (Value));  // eficaces -> eficaz
            End
            Else
              // ...//y + es -> ...//y
              If Ends (['', '', '', '', 'y', 'Y'], 'es') Then
              Begin
                // Exceptions (...//y + e)
                If Not CheckTruncate (['ayeayes', 'ensayes', 'hemates',
                'reensayes', 'yes']) Then
                  Result := ghDeleteRight (Value, 2);  // rubes -> rub
              End
              Else
                // ...Vowel a/e/i/o/u + ses -> ...Vowel + s
                If Ends (ghltESSimpleVowels, 'ses') Then
                Begin
                  // Exceptions (...vowel + se)
                  If Not CheckTruncate (['acuses', 'bases', 'ceses',
                  'chartreuses', 'clases', 'cmplases', 'desengrases',
                  'desfases', 'desgloses', 'desgrases', 'eclipses',
                  'elipses', 'embalses', 'endoses', 'engrases', 'enrases',
                  'envases', 'fases', 'frases', 'gloses', 'pases',
                  'rebalses', 'subclases', 'transvases', 'ucases']) Then
                  Begin
                    Result := ghDeleteRight (Value, 2);  // grises -> gris

                    If GHESSyllables (Result) > 1 Then
                      // Singular with accent (escoceses -> escocs)
                      ghAccentEnd (Result, 2);
                  End;
                End
                Else
                  // ...jes -> ...je
                  If Ends ([], 'jes') Then
                  Begin
                    // Exception (invariable form)
                    If SameText (Value, 'murajes') Then
                      Result := Value
                    Else
                      // Exceptions (...j)
                      If Not CheckTruncate (['almofrejes', 'almoradujes',
                      'bojes', 'cambujes', 'carcajes', 'errajes',
                      'maniblajes', 'relojes', 'trojes'], 2) Then
                        Result := ghDeleteRight (Value);  // ejes -> eje
                  End
                  Else
                    // ...Vowel i/o + des -> ...Vowel + de
                    If Ends (['i', 'I', 'o', 'O'], 'des') Then
                    Begin
                      // Exceptions (...vowel + d)
                      If Not CheckTruncate (['adalides', 'ardides',
                      'spides', 'cides', 'efodes', 'lides', 'quides',
                      'vides'], 2) And

                      // Exceptions (invariable form)
                      Not CheckInvariable (['alantoides', 'aracnoides',
                      'aritenoides', 'coroides', 'cricoides', 'deltoides',
                      'helicnides', 'herodes', 'mastoides', 'nomeolvides',
                      'paratiroides', 'tiroides', 'xifoides']) Then
                        // cspides -> cspide
                        Result := ghDeleteRight (Value);
                    End
                    Else
                      // ...Vowel a/e/i/o/u + xes -> ...Vowel + x
                      If Ends (ghltESSimpleVowels, 'xes') Then
                        // faxes -> fax
                        Result := ghDeleteRight (Value, 2)
                      Else
                        // ...Vowel a + s -> ...Vowel
                        If Ends (['a', 'A'], 's') Then
                        Begin
                          // Exceptions (invariable form)
                          If Not CheckInvariable (['abraxas', 'abrecartas',
                          'abrelatas', 'adivas', 'adobasillas',
                          'aguafiestas', 'ajaspajas', 'ajuagas',
                          'albricias', 'alcamonas', 'alias', 'ambas',
                          'andas', 'rganas', 'arras', 'arrebaaderas',
                          'artolas', 'atlas', 'atrapamoscas', 'baldragas',
                          'breas', 'braguillas', 'bramuras', 'bruselas',
                          'buleras', 'cachorreas', 'caderillas',
                          'calcillas', 'camahuas', 'caracas', 'cardinas',
                          'carisias', 'caristias', 'cascarrojas',
                          'castrapuercas', 'catacumbas', 'cataubas',
                          'cedras', 'ceideras', 'ceratias', 'cerbas',
                          'chaparreras', 'chotacabras', 'cierzas',
                          'contramangas', 'cortaplumas', 'cortaas',
                          'cosquillas', 'cuatrocientas', 'cuelgacapas',
                          'cuentagotas', 'descuernacabras',
                          'despabiladeras', 'despachaderas', 'despinzas',
                          'dimisorias', 'dormirlas', 'doscientas',
                          'dragaminas', 'ducas', 'echapellas',
                          'emborrachacabras', 'enagillas',
                          'entrecubiertas', 'espantamoscas', 'exequias',
                          'expensas', 'faldillas', 'finanzas',
                          'gallinejas', 'garbas', 'gemonas',
                          'guardabanderas', 'guardacostas', 'guardadamas',
                          'guardaespaldas', 'guardaguas', 'guardagujas',
                          'guardajoyas', 'guardapuntas', 'guardarruedas',
                          'halacabuyas', 'hurguillas', 'jamugas',
                          'jeremas', 'judas', 'lanzallamas', 'lavafrutas',
                          'lavazas', 'lechecillas', 'lemurias',
                          'limpiabotas', 'limpiachimeneas',
                          'limpiaparabrisas', 'limpiaplumas', 'limpiaas',
                          'litisexpensas', 'longuetas', 'marzas',
                          'matacabras', 'matacandelas', 'matamoscas',
                          'matasuegras', 'mecenas', 'mesas', 'milhojas',
                          'mondaorejas', 'montacargas', 'mulillas',
                          'natillas', 'nostras', 'novecientas', 'nupcias',
                          'ocenidas', 'ochocientas', 'orobias', 'pallas',
                          'panateneas', 'pncreas', 'pandectas',
                          'papamoscas', 'parabrisas', 'paracadas',
                          'paraguas', 'pasamontaas', 'pegamoscas',
                          'pesacartas', 'pisavas', 'portacartas',
                          'portaherramientas', 'portamantas',
                          'portamonedas', 'portanuevas', 'portaplumas',
                          'prensaestopas', 'quebrantaolas',
                          'quebrantapiedras', 'quinientas', 'quitamanchas',
                          'quitameriendas', 'rascatripas', 'recogemigas',
                          'retartalillas', 'revuelvepiedras',
                          'rompecabezas', 'rompeolas', 'sacabotas',
                          'sacabrocas', 'sacafilsticas', 'sacamanchas',
                          'sacamuelas', 'sacapelotas', 'sacapuntas',
                          'salvavidas', 'secafirmas', 'secundinas',
                          'seiscientas', 'serraduras', 'setecientas',
                          'smulas', 'tapajuntas', 'tenacillas',
                          'tirabotas', 'tirachinas', 'tiralneas',
                          'trabalenguas', 'tracias', 'trafalmejas',
                          'tragaperras', 'trescientas', 'valdepeas',
                          'veras', 'vierteaguas', 'vistillas']) Then
                            // casas -> casa
                            Result := ghDeleteRight (Value);
                        End
                        Else
                          // ...Vowel o + s -> ...Vowel
                          If Ends (['o', 'O'], 's') Then
                          Begin
                            // Exceptions (special plurals)
                            If Not CheckSpecial (['algunos', 'algn',
                            'esos', 'ese', 'estos', 'este', 'los',
                            'el']) And
                                              
                            // Exceptions (invariable form)
                            Not CheckInvariable (['aguamanos',
                            'aguavientos', 'ahorcaperros', 'albatros',
                            'alborotapueblos', 'ambos', 'ameos', 'amnios',
                            'aicos', 'apuracabos', 'argos',
                            'arrancaclavos', 'brtulos', 'bentos',
                            'besamanos', 'bicos', 'burdeos',
                            'buscarruidos', 'calientaplatos',
                            'calomelanos', 'canterios', 'caos', 'caporos',
                            'cariacos', 'carnios', 'castrapuercos',
                            'catacaldos', 'catervarios', 'caucos',
                            'cavaros', 'cazaclavos', 'cibarcos',
                            'ciensayos', 'comicios', 'contraarmios',
                            'contraveros', 'cortacallos', 'cortacircuitos',
                            'cortapicos', 'cortapuros', 'cosmos',
                            'cuajaenredos', 'cuatrocientos', 'cuentahlos',
                            'cuentakilmetros', 'cumpleaos',
                            'descalzaperros', 'diadelfos', 'doscientos',
                            'engaabobos', 'engaanecios', 'eros',
                            'escarbillos', 'escurreplatos',
                            'espantagustos', 'espantalobos',
                            'espantanublados', 'espantapjaros',
                            'extramuros', 'filopos', 'fondillos',
                            'galianos', 'greguescos', 'guardabarros',
                            'guardacartuchos', 'guardacuos',
                            'guardafrenos', 'hondillos', 'intramuros',
                            'lavamanos', 'lavaojos', 'lejsimos',
                            'lejitos', 'lejos', 'limpiabarros', 'loquios',
                            'matacallos', 'matagallegos', 'matamoros',
                            'matasellos', 'metemuertos', 'monadelfos',
                            'mondapozos', 'monospastos', 'novecientos',
                            'ochocientos', 'paralipmenos', 'pararrayos',
                            'pertrechos', 'psicos', 'poliadelfos',
                            'portalibros', 'portapliegos', 'pronaos',
                            'propleos', 'quebrantahuesos', 'quevedos',
                            'quinientos', 'quitamiedos', 'rascacielos',
                            'recogeabuelos', 'remediavagos',
                            'rompehielos', 'rompesacos', 'sacabocados',
                            'sacaclavos', 'sacacorchos', 'sacatrapos',
                            'salvabarros', 'seiscientos', 'sendos',
                            'setecientos', 'sinfisandrios', 'singensicos',
                            'tapacubos', 'taurios', 'ttanos',
                            'tocadiscos', 'trescientos', 'trotamundos',
                            'ultrapuertos', 'vilos']) Then
                              // cerros -> cerro
                              Result := ghDeleteRight (Value);
                          End
                          Else
                            // ...Vowel e/ + s -> ...Vowel
                            If Ends (ghltESEs, 's') Then
                            Begin
                              // Exceptions (special plurals)
                              If SameText (Value, 'albalaes') Or
                              SameText (Value, 'faralaes') Then
                                // albalaes -> albal, faralaes -> faral
                                Result := ghAccentedEnd (
                                  ghDeleteRight (Value, 2))
                              Else
                                // Exceptions (special plurals)
                                If Not CheckSpecial (['ferries', 'ferry',
                                'les', 'le', 'panties', 'panty', 'valses',
                                'vals']) And

                                // Exceptions (...)
                                Not CheckTruncate (['aes', 'lbumes',
                                'clubes', 'noes', 'oes', 'sndwiches',
                                'ttemes', 'yoes'], 2) And

                                // Exceptions (invariable form)
                                Not CheckInvariable (['alquermes',
                                'ambages', 'aries', 'arrastrapis',
                                'avisacoches', 'calientapis', 'caries',
                                'cayets', 'cazadotes', 'ciempis',
                                'contraaproches', 'diabetes', 'dragonites',
                                'enjuagadientes', 'etites', 'facies',
                                'hematites', 'jueves', 'lavacoches',
                                'limpiadientes', 'martes', 'milpis',
                                'mondadientes', 'parachoques', 'pasapurs',
                                'penates', 'quermes', 'quitanieves',
                                'saltamontes', 'sorites', 'tabes',
                                'termes', 'torques', 'trilobites',
                                'viernes']) Then
                                  // comits -> comit
                                  Result := ghDeleteRight (Value);
                            End 
                            Else
                              { ...a/e/i/o/u/p + sis ->
                                ...a/e/i/o/u/p + sis (invariable form) }
                              If Ends (ghltESSimpleVowels + ['p', 'P'],
                              'sis') Then
                              Begin
                                // Exceptions (...a/e/i/o/u/p + si )
                                If Not CheckTruncate (['diesis',
                                'jusis']) Then
                                  Result := Value;  // sinopsis -> sinopsis
                              End
                              Else
                                { ... + itis -> ... + itis (invariable
                                  form) }
                                If Ends ([], 'itis') Then
                                  Result := Value  // otitis -> otitis
                                Else
                                  // ...Vowel i/ + s -> ...Vowel
                                  If Ends (ghltESIs, 's') Then
                                  Begin
                                    // Exceptions (special plurals)
                                    If Not CheckSpecial (['cois', 'coy',
                                    'estis', 'estay', 'gais', 'gay',
                                    'guirigis', 'guirigay', 'jersis',
                                    'jersey', 'noris', 'noray', 'yoqueis',
                                    'yquey']) And
                                
                                    // Exceptions (invariable form)
                                    Not CheckInvariable (['acrpolis',
                                    'adonis', 'amarilis', 'anaptixis',
                                    'atrabilis', 'axis', 'bilis',
                                    'brindis', 'catarsis', 'centris',
                                    'chotis', 'cltoris', 'croquis',
                                    'cutis', 'dermis', 'diecisis',
                                    'dinornis', 'epidermis', 'epiglotis',
                                    'epistaxis', 'equis', 'filis',
                                    'finolis', 'frontis', 'gminis',
                                    'glacis', 'glotis', 'gneis', 'gratis',
                                    'ibis', 'intrngulis', 'iris',
                                    'satis', 'macis',  'mutis',
                                    'necrpolis', 'nipis', 'ostetis',
                                    'parataxis', 'prulis', 'pelvis',
                                    'pesquis', 'piscis', 'piscolabis',
                                    'praxis', 'profilaxis', 'pubis',
                                    'quepis', 'raquis', 'reis', 'sanguis',
                                    'seis', 'semis', 'sfilis', 'sintaxis',
                                    'tenis', 'tiquismiquis', 'tortcolis',
                                    'tris', 'unguis', 'uvetis', 'valais',
                                    'veintisis']) Then
                                      // confetis -> confeti
                                      Result := ghDeleteRight (Value);
                                  End
                                  Else
                                    // ...Vowel u/ + s -> ...Vowel
                                    If Ends (ghltESUs, 's') Then
                                    Begin
                                      // Exceptions (invariable form)
                                      If Not CheckInvariable (['ngelus',
                                      'cactus', 'crolus', 'corpus',
                                      'cristus', 'humus', 'idus', 'lapsus',
                                      'lupus', 'negus', 'mnibus',
                                      'rictus', 'sanctus', 'tifus',
                                      'venus', 'virus']) Then
                                        // mpetus -> mpetu
                                        Result := ghDeleteRight (Value);
                                    End
                                    Else
                                      // ...s -> ...
                                      If Ends ([], 's') Then
                                      Begin
                                        // Exceptions (invariable form)
                                        If Not CheckInvariable (['bceps',
                                        'cayaps', 'clrens', 'corps',
                                        'frceps', 'reps', 'trceps']) Then
                                          // mamuts -> mamut
                                          Result := ghDeleteRight (Value);
                                      End
                                      Else
                                        { Invariable form or unconsidered
                                          word (exequtur -> exequtur) }
                                        Result := Value;
  End;

  Function ghESA (Const Gender :TghGender; Const Plural :Boolean = False)
    :String;
  Const
    Words :Array [0..3] Of String = ('un', 'unos', 'una', 'unas');
  Begin
    { NOTE: Using an array constant parameter in a routine call requires
      less CPU instructions that using an open array constructor ("on the
      fly" array). }
    Result := ghGenderNumberStr (Words, Gender, Plural);
  End;

  Function ghESEnding (Const Gender :TghGender;
    Const Plural :Boolean = False) :String;
  Const
    Endings :Array [0..3] Of String = ('o', 'os', 'a', 'as');
  Begin
    { NOTE: Using an array constant parameter in a routine call requires
      less CPU instructions that using an open array constructor ("on the
      fly" array). }
    Result := ghGenderNumberStr (Endings, Gender, Plural);
  End;

  Function ghESEndingThat (Const Gender :TghGender;
    Const Plural :Boolean = False) :String;
  Begin
    Result := ghESEnding (Gender, Plural) + ' ' + ghESThat (Gender, Plural);
  End;

  Function ghESInLetters (Value :Int64) :String;
  Var
    Groups, Units :Integer;
    Negative :Boolean;

    Procedure Add (Const Str :String); Overload;
    Begin
      If Result <> '' Then
        Result := Result + ' ';

      Result := Result + Str;
    End;

    Procedure Add (Const Int :Integer); Overload;
    Begin
      Add (ghESInLetters (Int));
    End;
  Begin
    Result := '';
    Negative := Value < 0;
    Value := Abs (Value);

    If Value > 999999999999 Then  // Unsupported value
      Exit;

    Case Value Of
      // One-word number names
        0 : Result := 'cero';
        1 : Result := 'uno';
        2 : Result := 'dos';
        3 : Result := 'tres';
        4 : Result := 'cuatro';
        5 : Result := 'cinco';
        6 : Result := 'seis';
        7 : Result := 'siete';
        8 : Result := 'ocho';
        9 : Result := 'nueve';
       10 : Result := 'diez';
       11 : Result := 'once';
       12 : Result := 'doce';
       13 : Result := 'trece';
       14 : Result := 'catorce';
       15 : Result := 'quince';

       16..19, 21..29 :
       Begin
         If Value In [16..19] Then
           Result := 'dieci'
         Else
           Result := 'veinti';

         Case (Value Mod 10) Of
           2    : Result := Result + 'ds';
           3    : Result := Result + 'trs';
           6    : Result := Result + 'sis';
           Else
             Result := Result + ghESInLetters (Value Mod 10);
         End;
       End;

       20 : Result := 'veinte';

       30, 40, 50, 60, 70, 80, 90 :
       Begin
         Case Value Of
           30 : Result := 'trei';
           40 : Result := 'cuare';
           50 : Result := 'cincue';
           60 : Result := 'sese';
           70 : Result := 'sete';
           80 : Result := 'oche';
           90 : Result := 'nove';
         End;

         Result := Result + 'nta';
       End;

      100 : Result := 'cien';

      200, 300, 400, 500, 600, 700, 800, 900 :
      Begin
        Case Value Of
          200, 300, 400, 600, 800 :
            Result := ghESInLetters (Value Div 100) + 'c';

          500 : Result := 'quin';
          700 : Result := 'setec';
          900 : Result := 'novec';
        End;

        Result := Result + 'ientos';
      End;

      Else
      Begin
        Groups := Value Div 1000000;  // Millions

        If Groups > 0 Then
        Begin
          Result := ghESInLettersQuantity (Groups) + ' mill';

          If Groups = 1 Then
            Result := Result + 'n'
          Else
            Result := Result + 'ones';
        End;

        Groups := (Value Mod 1000000) Div 1000;  // Thousands

        If Groups > 0 Then
        Begin
          If Groups > 1 Then
            Add (ghESInLettersQuantity (Groups));

          Add ('mil');
        End;

        Groups := (Value Mod 1000) Div 100;  // Hundreds
        Units := Value Mod 100;  // Units (including the units in dozens)

        If Groups > 0 Then
        Begin
          Add (Groups * 100);

          If (Groups = 1) And (Units > 0) Then
            Result := Result + 'to';  // "cien" -> "ciento"
        End;

        // The numbers from 1 to 29 have one-word names
        If Units In [1..29] Then
          Add (Units)
        Else
          If Units > 0 Then
          Begin
            Add ((Units Div 10) * 10);  // Dozen 30..90
            Units := (Units Mod 10);  // Last digit

            // Separated units (74 -> "setenta y cuatro")
            If Units > 0 Then
              Add ('y ' + ghESInLetters (Units));
          End;
      End;
    End;

    If Negative Then
      Result := 'menos ' + Result;
  End;

  Function ghESInLettersQuantity (Const Value :Int64;
    Const Gender :TghGender = ghgnMasculine) :String;
  Var
    I :Integer;
  Begin
    Result := ghESInLetters (Value);

    If Gender = ghgnFeminine Then
      // Hundreds in feminine form (except hundreds of million)
      For I := Length (Result) DownTo 1 Do
        If Copy (Result, I - 2, 3) = 'tos' Then
          Result [I - 1] := 'a'  // "doscientos" -> "doscientas"
        Else
          // "milln"/"millones" reached
          If Copy (Result, I - 1, 2) = 'll' Then
            Break;

    If RightStr (Result, 2) = 'no' Then  // "uno" ending
    Begin
      // "sesenta y uno" -> "sesenta y un"
      Result := ghDeleteRight (Result);

      If Gender = ghgnFeminine Then
        Result := Result + 'a'  // "noventa y un" -> "noventa y una"
      Else
        If RightStr (Result, 3) = 'iun' Then  // "veintiun" ending
          Result [Length (Result) - 1] := '';  // "veintiun" -> "veintin"
    End;
  End;

  Function ghESIsDiphthong (Const Letter1, Letter2 :Char) :Boolean;
  Begin
    Result := (Letter1 <> Letter2) And  // Different vowels

    // Open vowel and unstressed closed vowel
    (((Letter1 In ghltESOpenVowels) And
    (Letter2 In ghltESSimpleClosedVowels)) Or

    // Unstressed closed vowel and open vowel
    ((Letter1 In ghltESSimpleClosedVowels) And
    (Letter2 In ghltESOpenVowels)) Or

    // Closed vowel and closed vowel
    ((Letter1 In ghltESClosedVowels) And (Letter2 In ghltESClosedVowels)));
  End;

  Function ghESThat (Const Gender :TghGender;
    Const Plural :Boolean = False) :String;
  Const
    Words :Array [0..3] Of String = ('ese', 'esos', 'esa', 'esas');
  Begin
    { NOTE: Using an array constant parameter in a routine call requires
      less CPU instructions that using an open array constructor ("on the
      fly" array). }
    Result := ghGenderNumberStr (Words, Gender, Plural);
  End;

  Function ghESThe (Const Gender :TghGender; Const Plural :Boolean = False)
    :String;
  Const
    Words :Array [0..3] Of String = ('el', 'los', 'la', 'las');
  Begin
    { NOTE: Using an array constant parameter in a routine call requires
      less CPU instructions that using an open array constructor ("on the
      fly" array). }
    Result := ghGenderNumberStr (Words, Gender, Plural);
  End;

  Function ghESToThe (Const Gender :TghGender;
    Const Plural :Boolean = False) :String;
  Const
    Words :Array [0..3] Of String = ('al', 'a los', 'a la', 'a las');
  Begin
    { NOTE: Using an array constant parameter in a routine call requires
      less CPU instructions that using an open array constructor ("on the
      fly" array). }
    Result := ghGenderNumberStr (Words, Gender, Plural);
  End;

Initialization
  GHESSingular := BasicSingular;
  GHESSingularWord := BasicSingularWord;
  GHESSyllables := BasicSyllables;

  If SysLocale.PriLangID = lang_Spanish Then
  Begin
    { NOTE: In Spanish, the names of days and months are written with
      lower case initial.  There is a possible related bug in Windows. }
    ghENSetLower (LongMonthNames);
    ghENSetLower (ShortMonthNames);
    ghENSetLower (LongDayNames);
    ghENSetLower (ShortDayNames);
  End;
End.

