{*************************************************************************}
{ GH Freebrary                                                            }
{                                                                         }
{ GHFOpenXMLSpreadsheet.pas - TghOpenXMLSpreadsheet class and             }
{                             TghOpenXMLSpreadsheetBook class unit.       }
{                                                                         }
{ This software is distributed under the open source BSD license.         }
{                                                                         }
{ Copyright (c) 2002-2014, Al Gonzalez (rescatandoadelphi.blogspot.com)   }
{ All rights reserved.                                                    }
{                                                                         }
{ Redistribution and use in source and binary forms, with or without      }
{ modification, are permitted provided that the following conditions are  }
{ met:                                                                    }
{                                                                         }
{   - Redistributions of source code must retain the above copyright      }
{     notice, this list of conditions and the following disclaimer.       }
{   - Redistributions in binary form must reproduce the above copyright   }
{     notice, this list of conditions and the following disclaimer in the }
{     documentation and/or other materials provided with the              }
{     distribution.                                                       }
{   - Neither the name of Al Gonzalez nor the names of his contributors   }
{     may be used to endorse or promote products derived from this        }
{     software without specific prior written permission.                 }
{                                                                         }
{ THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS }
{ IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED   }
{ TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A         }
{ PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT      }
{ HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,  }
{ SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT        }
{ LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,   }
{ DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY   }
{ THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT     }
{ (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE   }
{ OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.    }
{*************************************************************************}

{ Please contribute to improve this library.  I am not a good programmer,
  but good programmers know what Delphi is worth.

  A.G. }

Unit GHFOpenXMLSpreadsheet;  { OpenXML Spreadsheet }

Interface

  Uses
    GHFOpenXMLPart, GHFOpenXMLSpreadsheetStrs, GHFObjList;

  Type
    TghOpenXMLSpreadsheetBook = Class;

    { OpenXML Spreadsheet class }
    TghOpenXMLSpreadsheet = Class (TghOpenXMLPart)
      Protected
        { Fields }
        FBook :TghOpenXMLSpreadsheetBook;
        FStrs :TghOpenXMLSpreadsheetStrs;

        { Static methods }
        Function GetStrs :TghOpenXMLSpreadsheetStrs;

        { Virtual methods }
        Function GetValue (Const ACell :String) :Variant; Virtual;
      Public
        ID :Integer;
        
        Constructor Create (Const ABook :TghOpenXMLSpreadsheetBook;
          Const AContent :String = ''); Reintroduce; Overload; Virtual;
        Destructor Destroy; Override;

        { Static methods }
        Function Cell (Const Column :Char; Const Row :Integer) :OLEVariant;
          Overload;
        Function FindCell (Const Ref :String) :OLEVariant;
        Function Float (Const ACell :String) :Double; Overload;
        Function Float (Const Column :Char; Const Row :Integer) :Double;
          Overload;
        Function HasValue (Const ACell :String) :Boolean; Overload;
        Function HasValue (Const Column :Char; Const Row :Integer)
          :Boolean; Overload;
        Function Int (Const ACell :String) :Integer; Overload;
        Function Int (Const Column :Char; Const Row :Integer) :Integer;
          Overload;
        Function IsNull (Const ACell :String) :Boolean; Overload;
        Function IsNull (Const Column :Char; Const Row :Integer) :Boolean;
          Overload;
        Function Str (Const ACell :String) :String; Overload;
        Function Str (Const Column :Char; Const Row :Integer) :String;
          Overload;
        Function Value (Const Column :Char; Const Row :Integer) :Variant;

        { Virtual methods }
        Function Cell (Const Ref :String) :OLEVariant; Overload; Virtual;

        { Properties }
        Property Book :TghOpenXMLSpreadsheetBook Read FBook;
        Property Strs :TghOpenXMLSpreadsheetStrs Read GetStrs;
        Property Values [Const ACell :String] :Variant
          Read GetValue; Default;
    End;

    { OpenXML Spreadsheet Book class }
    TghOpenXMLSpreadsheetBook = Class (TghOpenXMLPart)
      Protected
        FSheets :TghObjList;
        FStrs :TghOpenXMLSpreadsheetStrs;

        { Static methods }
        Function GetStrs :TghOpenXMLSpreadsheetStrs;

        { Virtual methods }
        Function GetSheet (Const Name :String) :TghOpenXMLSpreadsheet;
          Virtual;
      Public
        Destructor Destroy; Override;

        { Static methods }
        Procedure SaveSheets;

        { Properties }
        Property Sheets [Const Name :String] :TghOpenXMLSpreadsheet
          Read GetSheet; Default;
        Property Strs :TghOpenXMLSpreadsheetStrs Read GetStrs;
    End;

Implementation

  Uses
    SysUtils, GHFRTL, Variants;

  { TghOpenXMLSpreadsheet }

  Constructor TghOpenXMLSpreadsheet.Create (
    Const ABook :TghOpenXMLSpreadsheetBook; Const AContent :String = '');
  Begin
    FBook := ABook;
    Create (AContent);
  End;

  Destructor TghOpenXMLSpreadsheet.Destroy;
  Begin
    // We free the owned string object (when FStrs <> Nil)
    FreeAndNil (FStrs);

    If (Book <> Nil) And (Book.FSheets <> Nil) Then
      Book.FSheets.Extract (Self);

    Inherited Destroy;
  End;

  { Protected static methods }

  Function TghOpenXMLSpreadsheet.GetStrs :TghOpenXMLSpreadsheetStrs;
  Begin
    If FStrs = Nil Then
      If Book <> Nil Then
      Begin
        Result := Book.Strs;
        Exit;
      End
      Else
        { sharedStrings.xml in the parent directory
          (...\xl\worksheets\sheet1.xml -> ...\xl\sharedStrings.xml) }
        FStrs := TghOpenXMLSpreadsheetStrs.CreateFromDir (ghDir (Path, 2));

    Result := FStrs;
  End;

  { Protected virtual methods }

  Function TghOpenXMLSpreadsheet.GetValue (Const ACell :String) :Variant;
  Var
    CellObj :OLEVariant;
  Begin
    CellObj := Cell (ACell);

    If VarIsClear (CellObj) Then
      Result := Null
    Else
    Begin
      Result := CellObj.SelectSingleNode ('_:v');

      If VarIsClear (Result) Then
        Result := Null
      Else
        // String type cell (value in sharedStrings.xml)
        If CellObj.GetAttribute ('t') = 's' Then
          Result := Strs [Result.Text]
        Else  // Other types use direct values
          Result := Result.Text;
    End;
  End;

  { Public static methods }

  Function TghOpenXMLSpreadsheet.Cell (Const Column :Char;
    Const Row :Integer) :OLEVariant;
  Begin
    Result := Cell (Column + IntToStr (Row));
  End;

  Function TghOpenXMLSpreadsheet.FindCell (Const Ref :String) :OLEVariant;
  Begin
    Result := Cell (Ref);

    If VarIsClear (Result) Then
      ghRaise ('Cell not found in spreadsheet ("%s").', [Ref]);
  End;

  Function TghOpenXMLSpreadsheet.Float (Const ACell :String) :Double;
  Begin
    Result := ghFloat (Values [ACell]);
  End;

  Function TghOpenXMLSpreadsheet.Float (Const Column :Char;
    Const Row :Integer) :Double;
  Begin
    Result := ghFloat (Value (Column, Row));
  End;

  Function TghOpenXMLSpreadsheet.HasValue (Const ACell :String) :Boolean;
  Begin
    Result := Not VarIsNull (Values [ACell]);
  End;

  Function TghOpenXMLSpreadsheet.HasValue (Const Column :Char;
    Const Row :Integer) :Boolean;
  Begin
    Result := Not VarIsNull (Value (Column, Row));
  End;

  Function TghOpenXMLSpreadsheet.Int (Const ACell :String) :Integer;
  Begin
    Result := ghInt (Values [ACell]);
  End;

  Function TghOpenXMLSpreadsheet.Int (Const Column :Char;
    Const Row :Integer) :Integer;
  Begin
    Result := ghInt (Value (Column, Row));
  End;

  Function TghOpenXMLSpreadsheet.IsNull (Const ACell :String) :Boolean;
  Begin
    Result := VarIsNull (Values [ACell]);
  End;

  Function TghOpenXMLSpreadsheet.IsNull (Const Column :Char;
    Const Row :Integer) :Boolean;
  Begin
    Result := VarIsNull (Value (Column, Row));
  End;

  Function TghOpenXMLSpreadsheet.Str (Const ACell :String) :String;
  Begin
    Result := VarToStr (Values [ACell]);
  End;

  Function TghOpenXMLSpreadsheet.Str (Const Column :Char;
    Const Row :Integer) :String;
  Begin
    Result := VarToStr (Value (Column, Row));
  End;

  Function TghOpenXMLSpreadsheet.Value (Const Column :Char;
    Const Row :Integer) :Variant;
  Begin
    Result := Values [Column + IntToStr (Row)];
  End;

  { Public virtual methods }

  Function TghOpenXMLSpreadsheet.Cell (Const Ref :String) :OLEVariant;
  Begin
    { First cell (/worksheet/sheetData/row/c) with the r attribute equal to
      Ref }
    Result := Node ('/_:worksheet/_:sheetData/_:row/_:c[@r="%s"]',
      [UpperCase (Ref)]);
  End;

  { TghOpenXMLSpreadsheetBook }

  Destructor TghOpenXMLSpreadsheetBook.Destroy;
  Begin
    FreeAndNil (FSheets);
    FreeAndNil (FStrs);
    Inherited Destroy;
  End;

  { Protected static methods }

  Function TghOpenXMLSpreadsheetBook.GetStrs :TghOpenXMLSpreadsheetStrs;
  Begin
    If FStrs = Nil Then
      // sharedStrings.xml in the same directory that workbook.xml
      FStrs := TghOpenXMLSpreadsheetStrs.CreateFromDir (ghDir (Path));

    Result := FStrs
  End;

  { Protected virtual methods }

  Function TghOpenXMLSpreadsheetBook.GetSheet (Const Name :String)
    :TghOpenXMLSpreadsheet;
  Var
    ANode :OLEVariant;
    I, ID :Integer;
  Begin
    { First sheet entry (/workbook/sheets/sheet) with the name attribute
      equal to Name }
    ANode := FindNode ('/_:workbook/_:sheets/_:sheet[@name="%s"]', [Name]);

    ID := ANode.GetAttribute ('sheetId');

    If FSheets <> Nil Then
      For I := 0 To FSheets.Count - 1 Do
        If (FSheets [I] As TghOpenXMLSpreadsheet).ID = ID Then
        Begin
          Result := TghOpenXMLSpreadsheet (FSheets [I]);
          Exit;
        End;

    Result := TghOpenXMLSpreadsheet.Create (Self,
      Rels.ElementTargetPath (ANode));

    Try
      Result.ID := ID;
      
      If FSheets = Nil Then
        FSheets := TghObjList.Create;

      FSheets.Add (Result);
    Except
      Result.Free;
      Raise;
    End;
  End;

  { Public static methods }

  Procedure TghOpenXMLSpreadsheetBook.SaveSheets;
  Var
    I :Integer;
  Begin
    If FSheets <> Nil Then
      For I := 0 To FSheets.Count - 1 Do
        (FSheets [I] As TghOpenXMLSpreadsheet).Save;
  End;

End.
