{*************************************************************************}
{ GH Freebrary                                                            }
{                                                                         }
{ GHFSQLQuery.pas - TghSQLQuery class unit.                               }
{                                                                         }
{ This software is distributed under the open source BSD license.         }
{                                                                         }
{ Copyright (c) 2002-2014, Al Gonzalez (rescatandoadelphi.blogspot.com)   }
{ All rights reserved.                                                    }
{                                                                         }
{ Redistribution and use in source and binary forms, with or without      }
{ modification, are permitted provided that the following conditions are  }
{ met:                                                                    }
{                                                                         }
{   - Redistributions of source code must retain the above copyright      }
{     notice, this list of conditions and the following disclaimer.       }
{   - Redistributions in binary form must reproduce the above copyright   }
{     notice, this list of conditions and the following disclaimer in the }
{     documentation and/or other materials provided with the              }
{     distribution.                                                       }
{   - Neither the name of Al Gonzalez nor the names of his contributors   }
{     may be used to endorse or promote products derived from this        }
{     software without specific prior written permission.                 }
{                                                                         }
{ THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS }
{ IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED   }
{ TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A         }
{ PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT      }
{ HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,  }
{ SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT        }
{ LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,   }
{ DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY   }
{ THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT     }
{ (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE   }
{ OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.    }
{*************************************************************************}

{ Please contribute to improve this library.  I am not a good programmer,
  but good programmers know what Delphi is worth.

  A.G. }

Unit GHFSQLQuery;  { SQL Query }

Interface

  Uses
    SQLExpr, DB, Classes;

  Type
    { SQL Query class }
    TghSQLQuery = Class (TSQLQuery)
      Protected
        { Fields }
        FClearUnprovidedParams :Boolean;
        FStoreActive :Boolean;

        { Overridden methods }
        Procedure DefineProperties (Filer :TFiler); Override;
        Procedure PSSetParams (AParams :TParams); Override;
      Published
        { Properties }
        Property Active Stored FStoreActive;
        Property ClearUnprovidedParams :Boolean Read FClearUnprovidedParams
          Write FClearUnprovidedParams Default False;

        { Workaround for Delphi quality report 7991 (access violations with
          subcomponents in visual inheritance), useful if the
          InnerDataSetClass property of a TghDataSetProvider component is
          'TghSQLQuery'.  We store the properties manually in the
          DefineProperties method. }
        Property Params Stored False;
        Property SchemaName Stored False;
        Property SQLConnection Stored False;

        Property StoreActive :Boolean Read FStoreActive Write FStoreActive
          Default False;
    End;

Implementation

  Uses
    TypInfo;

  { TghSQLQuery }

  { Protected overridden methods }

    Type
      TWriterAccess = Class (TWriter);
  Procedure TghSQLQuery.DefineProperties (Filer :TFiler);
  Begin
    Inherited DefineProperties (Filer);

    { Workaround for Delphi quality report 7991 (see the property
      declarations) }
    If Filer Is TWriter Then
    Begin
      TWriterAccess (Filer).WriteProperty (
        Self, GetPropInfo (Self, 'Params'));
      TWriterAccess (Filer).WriteProperty (
        Self, GetPropInfo (Self, 'SchemaName'));
      TWriterAccess (Filer).WriteProperty (
        Self, GetPropInfo (Self, 'SQLConnection'));
    End;
  End;

  Procedure TghSQLQuery.PSSetParams (AParams :TParams);
  Var
    I :Integer;
    NewParams :TParams;
    Param :TParam;
  Begin
    { We assign the values only and clear parameters not specified (when
      ClearUnprovidedParams is True), instead of replace the complete
      collection }

    NewParams := TParams.Create;

    Try
      NewParams.Assign (Params);

      For I := 0 To NewParams.Count - 1 Do
      Begin
        Param := AParams.FindParam (NewParams [I].Name);

        If Param <> Nil Then
          NewParams [I].Value := Param.Value
        Else
          If ClearUnprovidedParams Then
            NewParams [I].Clear;
      End;

      Inherited PSSetParams (NewParams);
    Finally
      NewParams.Free;
    End;
  End;
  
End.
 