{*************************************************************************}
{ GH Freebrary                                                            }
{                                                                         }
{ GHFStrList.pas - TghStrList class unit.                                 }
{                                                                         }
{ This software is distributed under the open source BSD license.         }
{                                                                         }
{ Copyright (c) 2002-2014, Al Gonzalez (rescatandoadelphi.blogspot.com)   }
{ All rights reserved.                                                    }
{                                                                         }
{ Redistribution and use in source and binary forms, with or without      }
{ modification, are permitted provided that the following conditions are  }
{ met:                                                                    }
{                                                                         }
{   - Redistributions of source code must retain the above copyright      }
{     notice, this list of conditions and the following disclaimer.       }
{   - Redistributions in binary form must reproduce the above copyright   }
{     notice, this list of conditions and the following disclaimer in the }
{     documentation and/or other materials provided with the              }
{     distribution.                                                       }
{   - Neither the name of Al Gonzalez nor the names of his contributors   }
{     may be used to endorse or promote products derived from this        }
{     software without specific prior written permission.                 }
{                                                                         }
{ THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS }
{ IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED   }
{ TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A         }
{ PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT      }
{ HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,  }
{ SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT        }
{ LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,   }
{ DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY   }
{ THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT     }
{ (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE   }
{ OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.    }
{*************************************************************************}

{ Please contribute to improve this library.  I am not a good programmer,
  but good programmers know what Delphi is worth.

  A.G. }

Unit GHFStrList;  { String List }

Interface

  Uses
    Classes, GHFRTL;

  Type
    { String List class }
    TghStrList = Class (TStringList)
      Protected
        { Static methods }
        Procedure SetCount (Const Value :Integer);
      Public
        Constructor Create (Const ASorted :Boolean); Overload; Virtual;

        { Static methods }
        Procedure DeleteAfter (Const Value :String);
        Procedure DeleteFrom (Const Value :String);
        Procedure DeleteLast;
        Function FreeObj (Const Index :Integer) :Boolean;
        Procedure FreeObjs;
        Function IndexPos (Const SubStr :String;
          Const StartChrIndex :Integer = 1) :TghTwoInts;
        Function Last :String;

        { Properties }
        Property Count Write SetCount;
    End;

Implementation

  { TghStrList }

  Constructor TghStrList.Create (Const ASorted :Boolean);
  Begin
    Inherited Create;
    Sorted := ASorted;
  End;

  { Protected static methods }

  Procedure TghStrList.SetCount (Const Value :Integer);
  Var
    I :Integer;
  Begin
    I := Value - Count;
    BeginUpdate;

    Try
      If I > 0 Then
        For I := 1 To I Do
          Add ('')
      Else
        For I := -1 DownTo I Do
          Delete (Count - 1);
    Finally
      EndUpdate;
    End;
  End;

  { Public static methods }

  Procedure TghStrList.DeleteAfter (Const Value :String);
  Var
    I :Integer;
  Begin
    I := IndexOf (Value);

    If I > -1 Then
      Count := I + 1;
  End;

  Procedure TghStrList.DeleteFrom (Const Value :String);
  Var
    I :Integer;
  Begin
    I := IndexOf (Value);

    If I > -1 Then
      Count := I;
  End;

  Procedure TghStrList.DeleteLast;
  Begin
    If Count > 0 Then
      Delete (Count - 1);
  End;

  Function TghStrList.FreeObj (Const Index :Integer) :Boolean;
  Var
    Obj :TObject;
  Begin
    Obj := Objects [Index];
    Result := Obj <> Nil;

    If Result Then
    Begin
      Objects [Index] := Nil;
      Obj.Free;
    End;
  End;

  Procedure TghStrList.FreeObjs;
  Var
    I :Integer;
  Begin
    I := 0;

    While I < Count Do
      If FreeObj (I) Then
        I := 0
      Else
        Inc (I);
  End;

  Function TghStrList.IndexPos (Const SubStr :String;
    Const StartChrIndex :Integer = 1) :TghTwoInts;
  Var
    I :Integer;
  Begin
    For I := 0 To Count - 1 Do
    Begin
      Result.LowValue := ghPos (Strings [I], SubStr, StartChrIndex,
        TghStrCompare (Byte (ghscLocaleAnyCase) - Byte (CaseSensitive)));

      If Result.LowValue > 0 Then
      Begin
        Result.HighValue := I;
        Exit;
      End;
    End;

    Result.HighValue := -1;
    Result.LowValue := 0;
  End;

  Function TghStrList.Last :String;
  Begin
    Result := Strings [Count - 1];
  End;

End.
