{*************************************************************************}
{ GH Freebrary                                                            }
{                                                                         }
{ GHFVCL.pas - Unit of VCL related elements.                              }
{                                                                         }
{ This software is distributed under the open source BSD license.         }
{                                                                         }
{ Copyright (c) 2002-2014, Al Gonzalez (rescatandoadelphi.blogspot.com)   }
{ All rights reserved.                                                    }
{                                                                         }
{ Redistribution and use in source and binary forms, with or without      }
{ modification, are permitted provided that the following conditions are  }
{ met:                                                                    }
{                                                                         }
{   - Redistributions of source code must retain the above copyright      }
{     notice, this list of conditions and the following disclaimer.       }
{   - Redistributions in binary form must reproduce the above copyright   }
{     notice, this list of conditions and the following disclaimer in the }
{     documentation and/or other materials provided with the              }
{     distribution.                                                       }
{   - Neither the name of Al Gonzalez nor the names of his contributors   }
{     may be used to endorse or promote products derived from this        }
{     software without specific prior written permission.                 }
{                                                                         }
{ THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS }
{ IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED   }
{ TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A         }
{ PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT      }
{ HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,  }
{ SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT        }
{ LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,   }
{ DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY   }
{ THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT     }
{ (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE   }
{ OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.    }
{*************************************************************************}

{ Please contribute to improve this library.  I am not a good programmer,
  but good programmers know what Delphi is worth.

  A.G. }

Unit GHFVCL;  { VCL }

Interface

  Uses
    DB, Classes, Controls, DSIntf, GHFRTL, StdCtrls, Windows, Forms,
    Dialogs;

  Const
    { DataSnap Record attributes }
    ghdrChanges = dsRecDeleted Or dsRecNew Or dsRecModified;

    { TDataSetState sets }

    ghdsBrowses           = [dsBrowse, dsBlockRead];

    // Calculations
    ghdsCalcs             = [dsCalcFields, dsInternalCalc];

    ghdsEdits             = [dsEdit, dsInsert];
    ghdsBasics            = [dsInactive, dsBrowse] + ghdsEdits;
    ghdsBrowsesEdits      = ghdsBrowses + ghdsEdits;
    ghdsBrowsesEditsCalcs = ghdsBrowsesEdits + ghdsCalcs;
    ghdsEditCalcs         = [dsEdit] + ghdsCalcs;
    ghdsEditsCalcs        = ghdsEdits + ghdsCalcs;

    { TFieldType sets }
    
    ghftAll      = [Low (TFieldType)..High (TFieldType)];
    ghftKnowns   = ghftAll - [ftUnknown];

    // Integers
    ghftInts     = [ftSmallint, ftInteger, ftWord, ftAutoInc, ftLargeInt];
    
    ghftFloats   = [ftFloat, ftCurrency, ftBCD, ftFMTBcd];
    ghftNumerics = ghftInts + ghftFloats;

    { Update Status }
    ghusChanges = [usModified, usInserted, usDeleted];

  Type
    { Data Set Class type }
    TDataSetClass = Class Of TDataSet;

    { Before Continue Field Event type }
    TghBeforeContinueFieldEvent = Procedure (Field :TField;
      Var AContinue :Boolean) Of Object;

    { Change Update Status type }
    TghChangeUpdateStatus = usModified..usDeleted;

    { Change Update Status Set type }
    TghChangeUpdateStatusSet = Set Of TghChangeUpdateStatus;

    { Clone Data Set Event type }
    TghCloneDataSetEvent = Procedure (Sender :TObject;
      DataSet :TDataSet; Var AClass :TDataSetClass;
      Out Clone :TDataSet) Of Object;

    { Data Events type }
    TghDataEvents = Set Of TDataEvent;

    { Data Set Filtering type }
    TghDataSetFiltering = Packed Record
      Active :Boolean;
      Expr :String;
      Event :TFilterRecordEvent;
      Bookmark :TBookmark;
    End;

    TghFieldAfterChangeEvent = Procedure (Field :TField;
      Const PrevValue :Variant) Of Object;

    TghFieldBeforeChangeEvent = Procedure (Field :TField;
      Const NewValue :Variant) Of Object;

    { Field Types type }
    TghFieldTypes = Set Of TFieldType;

    { ghmsCloseModal type }
    TghmsCloseModal = Packed Record
      ID :Cardinal;
      ModalResult :TModalResult;
      Unused :LongInt;
      Result :LongInt;
    End;

    PCustomForm = ^TCustomForm;

  Var
    { Message Icon Names }
    GHMsgIconNames :Array [TMsgDlgType] Of String =
      (ghinWarning, ghinStop, ghinInfo, ghinQuestion, '');

  { Functions and procedures }

  { Able? }
  Function ghAble (Const Control :TControl) :Boolean;

  { Abort }
  Procedure ghAbort (Const Sound :Integer = mb_IconStop;
    Const FocusControl :TWinControl = Nil); Overload;

  { Abort }
  Procedure ghAbort (Const FocusControl :TWinControl); Overload;

  { Abort }
  Procedure ghAbort (Const Caption, Text :String;
    Const FocusControl :TWinControl = Nil); Overload;

  { Abort }
  Procedure ghAbort (Const Text :String;
    Const FocusControl :TWinControl = Nil); Overload;

  { Abort if No }
  Procedure ghAbortNo (Const Caption, Text :String;
    Const FocusControl :TWinControl = Nil); Overload;

  { Abort if No }
  Procedure ghAbortNo (Const Text :String;
    Const FocusControl :TWinControl = Nil); Overload;

  { Abort if Yes }
  Procedure ghAbortYes (Const Caption, Text :String;
    Const FocusControl :TWinControl = Nil); Overload;

  { Abort if Yes }
  Procedure ghAbortYes (Const Text :String;
    Const FocusControl :TWinControl = Nil); Overload;

  { Ask }
  Function ghAsk (Const Caption, Text :String;
    Const Options :Integer = 0) :Integer; Overload;

  { Ask }
  Function ghAsk (Const Text :String; Const Options :Integer = 0) :Integer;
    Overload;

  { Ask Retry }
  Function ghAskRetry (Const Caption, Text :String;
    Const Options :Integer = 0) :Integer; Overload;

  { Ask Retry }
  Function ghAskRetry (Const Text :String; Const Options :Integer = 0)
    :Integer; Overload;

  { Ask Retry Abort }
  Function ghAskRetryAbort (Const Caption, Text :String;
    Const Options :Integer = 0) :Integer; Overload;

  { Ask Retry Abort }
  Function ghAskRetryAbort (Const Text :String; Const Options :Integer = 0)
    :Integer; Overload;

  { Can Focus? }
  Function ghCanFocus (Const Control :TWinControl;
    Const UnableForm :PCustomForm = Nil) :Boolean;

  { Check User Level }
  Function ghCheckUserLevel (Const Min :Integer) :Integer; Overload;

  { Check User Level }
  Function ghCheckUserLevel (Const Min :Integer;
    Const ErrorCaption, ErrorText :String) :Integer; Overload;

  { Check User Level }
  Function ghCheckUserLevel (Const Min :Integer; Const ErrorText :String)
    :Integer; Overload;

  { Clear Parameters }
  Procedure ghClearParams (Const Params :TParams);

  { Clone Data Set }
  Function ghCloneDataSet (Const DataSet :TDataSet;
    Const AClass :TDataSetClass; Const Owner :TComponent = Nil;
    ExcludedProps :String = 'Active'; Const IncludeFields :Boolean = True;
    Const ExcludedFieldProps :String = '') :TDataSet; Overload;

  { Clone Data Set }
  Function ghCloneDataSet (Const DataSet :TDataSet;
    Const Owner :TComponent = Nil; Const ExcludedProps :String = 'Active';
    Const IncludeFields :Boolean = True;
    Const ExcludedFieldProps :String = '') :TDataSet; Overload;

  { Container }
  Function ghContainer (Const Control :TControl;
    Const AClass :TWinControlClass) :TWinControl; Overload;

  { Container }
  Function ghContainer (Const AClass :TWinControlClass) :TWinControl;
    Overload;

  { Copy Data Set Properties }
  Procedure ghCopyDataSetProps (Const Source, Dest :TDataSet;
    Const Excludeds :String = ''; Const IncludeFields :Boolean = True;
    Const ExcludedFieldProps :String = '');

  { Copy Fields }
  Procedure ghCopyFields (Const Source, Dest :TDataSet;
    Const ExcludedProps :String = '');

  { DataSnap Attribute }
  Function ghDataSnapAttr (Const UpdateStatus :TUpdateStatus) :Integer;

  { DataSnap Attributes }
  Function ghDataSnapAttrs (Const UpdateStatus :TUpdateStatusSet) :Integer;

  { DataSnap Update Status }
  Function ghDataSnapUpdateStatus (Const Attr :Integer) :TUpdateStatus;

  { DataSnap Update Status Set }
  Function ghDataSnapUpdateStatusSet (Const Attrs :Integer)
    :TUpdateStatusSet;

  { Destroying? }
  Function ghDestroying (Const Control :TControl) :Boolean;

  { Dialog Type }
  Function ghDialogType (Const IconFlag :Integer) :TMsgDlgType;

  { Equal Fields? }
  Function ghEqualFields (Const List1, List2 :TList) :Boolean;

  { Field Count }
  Function ghFieldCount (Const Fields :String) :Integer;

  { Field Type }
  Function ghFieldType (Const Name :String) :TFieldType;

  { Field Type Extended }
  Function ghFieldTypeEx (Const AType :String) :TFieldType;

  { Field Type Integer }
  Function ghFieldTypeInt (Const Name :String) :Integer;

  { Filtering }
  Function ghFiltering (Const DataSet :TDataSet) :TghDataSetFiltering;

  { Focus }
  Function ghFocus (Const Control :TWinControl) :Boolean;

  { Inform }
  Function ghInform (Const Caption, Text :String;
    Const Options :Integer = 0) :Integer; Overload;

  { Inform }
  Function ghInform (Const Text :String; Const Options :Integer = 0)
    :Integer; Overload;

  { Input Parameter }
  Function ghInputParam (Const Params :TParams; Index :Integer = 0)
    :TParam;

  { Input Parameter Count }
  Function ghInputParamCount (Const Params :TParams) :Integer;

  { Is? }
  Function ghIs (Const Window :THandle; Const AClass :TWinControlClass)
    :Boolean;

  { Is Button? }
  Function ghIsButton (Const Window :THandle) :Boolean;

  { Make Able }
  Procedure ghMakeAble (Const Control :TControl);

  { Master Link }
  Function ghMasterLink (Const ADataSet :TDataSet) :TMasterDataLink;

  { Modal Level }
  Function ghModalLevel :Integer;

  { Message Icon }
  Function ghMsgIcon (Const Options :Integer; Var Icon :String) :Integer;

  { Message Icon Check }
  Function ghMsgIconCheck (Const Options :Integer; Var Icon :String)
    :Integer;

  { Message Icon Name }
  Function ghMsgIconName (Const Flag :Integer;
    Const ValidateFlag :Boolean = True) :String;

  { Message Box Parameters }
  Function ghMsgBoxParams (Const Options :Integer = 0) :TMsgBoxParams;
    Overload;

  { Message Box Parameters }
  Function ghMsgBoxParams (Const Caption, Text :String;
    Const Options :Integer = 0) :TMsgBoxParams; Overload;

  { Message Box Parameters }
  Function ghMsgBoxParams (Const Text :String; Const Options :Integer = 0)
    :TMsgBoxParams; Overload;

  { Message Box Parameters }
  Function ghMsgBoxParams (Const Caption, Text :String;
    Const Options :Integer; Const Icon :PChar;
    Const IconModule :THandle = 0) :TMsgBoxParams; Overload;

  { Message Box Parameters }
  Function ghMsgBoxParams (Const Text :String; Const Options :Integer;
    Const Icon :PChar; Const IconModule :THandle = 0) :TMsgBoxParams;
    Overload;

  { Message Box Parameters }
  Function ghMsgBoxParams (Const Caption, Text :String;
    Const Icon :PChar; Const IconModule :THandle = 0) :TMsgBoxParams;
    Overload;

  { Message Box Parameters }
  Function ghMsgBoxParams (Const Options :Integer; Const Icon :PChar;
    Const ResourceModule :THandle = 0) :TMsgBoxParams; Overload;

  { Message Box Parameters Marks }
  Function ghMsgBoxParamsMarks (Const Caption, Text :String;
    Const Options :Integer) :TMsgBoxParams;

  { Output Parameter }
  Function ghOutputParam (Const Params :TParams; Index :Integer = 0)
    :TParam;

  { Output Parameter Count }
  Function ghOutputParamCount (Const Params :TParams) :Integer;

  { Press Character }
  Function ghPressChr (Const Control :TControl; Const Value :Char;
    Const KeyData :Integer = 0) :Integer;

  { Root Control }
  Function ghRootControl (Const Control :TControl) :TControl;

  { Set Cursor }
  Function ghSetCursor (Const Value :TCursor) :TCursor;

  { Show Message }
  Function ghShowMsg (Const Params :TMsgBoxParams) :Integer; Overload;

  { Show Message }
  Function ghShowMsg (Const Params :TMsgBoxParams; Const Sound :Integer)
    :Integer; Overload;
    
  { Show Message }
  Function ghShowMsg (Const Caption, Text :String;
    Const Options :Integer = 0; Icon :String = '') :Integer; Overload;

  { Show Message }
  Function ghShowMsg (Const Text :String; Const Options :Integer = 0;
    Const Icon :String = '') :Integer; Overload;

  { Show Message }
  Function ghShowMsg (Const Caption, Text, Icon :String) :Integer;
    Overload;

  { Show Message Icon }
  Function ghShowMsgIcon (Const Text, Icon :String) :Integer; Overload;

  { Show Stop }
  Function ghShowStop (Const Caption, Text :String;
    Const Options :Integer = 0; Const FocusControl :TWinControl = Nil)
    :Integer; Overload;

  { Show Stop }
  Function ghShowStop (Const Text :String; Const Options :Integer = 0;
    Const FocusControl :TWinControl = Nil) :Integer; Overload;

  { Show Stop }
  Function ghShowStop (Const Caption, Text :String;
    Const FocusControl :TWinControl) :Integer; Overload;

  { Show Stop }
  Function ghShowStop (Const Text :String; Const FocusControl :TWinControl)
    :Integer; Overload;

  { SQL }
  Function ghSQL (Const DataSet :TDataSet) :String;

  { SQL List }
  Function ghSQLList (Const DataSet :TDataSet) :TStrings;

  { SQL Table }
  Function ghSQLTable (Const DataSet :TDataSet) :String;

  { Table }
  Function ghTable (Const DataSet :TDataSet) :String;

  { Title }
  Function ghTitle (Const Value :String) :String;

  { Title PChar }
  Function ghTitlePChr :PChar; Overload;

  { Title PChar }
  Function ghTitlePChr (Const Value :String) :PChar; Overload;

  { Title Pointer }
  Function ghTitlePtr :Pointer; Overload;

  { Title Pointer }
  Function ghTitlePtr (Const Value :String) :Pointer; Overload;

  { Usable? }
  Function ghUsable (Const Control :TControl) :Boolean;

  { Want Enter? }
  Function ghWantEnter (Const Window :THandle) :Boolean;

  { Warn }
  Function ghWarn (Const Caption, Text :String;
    Const Options :Integer = 0) :Integer; Overload;

  { Warn }
  Function ghWarn (Const Text :String; Const Options :Integer = 0)
    :Integer; Overload;

Implementation

  Uses
    DBCommon, TypInfo, SysUtils, Math, Messages;

  Var
    AppTitle :String;

  { Functions and procedures }

  Function ghAble (Const Control :TControl) :Boolean;
  Begin
    Result := Control.Visible And Control.Enabled;
  End;

  Procedure ghAbort (Const Sound :Integer = mb_IconStop;
    Const FocusControl :TWinControl = Nil); Overload;
  Begin
    MessageBeep (Sound);

    If FocusControl <> Nil Then
      ghFocus (FocusControl);

    Abort;
  End;

  Procedure ghAbort (Const FocusControl :TWinControl); Overload;
  Begin
    ghAbort (mb_IconStop, FocusControl);
  End;

  Procedure ghAbort (Const Caption, Text :String;
    Const FocusControl :TWinControl = Nil); Overload;
  Begin
    ghShowStop (Caption, Text, FocusControl);
    Abort;
  End;

  Procedure ghAbort (Const Text :String;
    Const FocusControl :TWinControl = Nil); Overload;
  Begin
    ghAbort ('', Text, FocusControl);
  End;

  Procedure ghAbortNo (Const Caption, Text :String;
    Const FocusControl :TWinControl = Nil); Overload;
  Begin
    If ghAsk (Caption, Text) = IDNo Then
    Begin
      If FocusControl <> Nil Then
        ghFocus (FocusControl);

      Abort;
    End;
  End;

  Procedure ghAbortNo (Const Text :String;
    Const FocusControl :TWinControl = Nil); Overload;
  Begin
    ghAbortNo ('', Text, FocusControl);
  End;

  Procedure ghAbortYes (Const Caption, Text :String;
    Const FocusControl :TWinControl = Nil); Overload;
  Begin
    If ghAsk (Caption, Text) = IDYes Then
    Begin
      If FocusControl <> Nil Then
        ghFocus (FocusControl);

      Abort;
    End;
  End;

  Procedure ghAbortYes (Const Text :String;
    Const FocusControl :TWinControl = Nil); Overload;
  Begin
    ghAbortYes ('', Text, FocusControl);
  End;

  Function ghAsk (Const Caption, Text :String;
    Const Options :Integer = 0) :Integer; Overload;
  Begin
    Result := ghShowMsg (Caption, Text,
      ghCheckFlagButtonIcon (Options, mb_YesNo, mb_IconQuestion));
  End;

  Function ghAsk (Const Text :String; Const Options :Integer = 0) :Integer;
    Overload;
  Begin
    Result := ghAsk ('', Text, Options);
  End;

  Function ghAskRetry (Const Caption, Text :String;
    Const Options :Integer = 0) :Integer; Overload;
  Begin
    Result := ghShowMsg (Caption, Text, ghCheckFlagIcon (
      ghFlagButton (Options, mb_RetryCancel), mb_IconStop));
  End;

  Function ghAskRetry (Const Text :String; Const Options :Integer = 0)
    :Integer; Overload;
  Begin
    Result := ghAskRetry ('', Text, Options);
  End;

  Function ghAskRetryAbort (Const Caption, Text :String;
    Const Options :Integer = 0) :Integer; Overload;
  Begin
    Result := ghShowMsg (Caption, Text, ghCheckFlagIcon (
      ghFlagButton (Options, mb_AbortRetryIgnore), mb_IconStop));
  End;

  Function ghAskRetryAbort (Const Text :String; Const Options :Integer = 0)
    :Integer; Overload;
  Begin
    Result := ghAskRetryAbort ('', Text, Options);
  End;

  Function ghCanFocus (Const Control :TWinControl;
    Const UnableForm :PCustomForm = Nil) :Boolean;
  Var
    Form :TCustomForm;
  Begin
    Result := False;

    If Control.CanFocus And Not ghDestroying (Control) Then
    Begin
      Form := GetParentForm (Control);

      If Form = Nil Then
        Result := True
      Else
        If (Form = Control) Or Not ghDestroying (Form) Then
          If ghAble (Form) Then
            Result := True
          Else                          
          Begin
            ghSetPtr (UnableForm, Form);
            Exit;
          End;
    End;

    ghSetPtr (UnableForm);
  End;

  Function ghCheckUserLevel (Const Min :Integer) :Integer; Overload;
  Begin
    Result := GHURCurrent.Level;

    If Result < Min Then
      ghAbort;
  End;

  Function ghCheckUserLevel (Const Min :Integer;
    Const ErrorCaption, ErrorText :String) :Integer; Overload;
  Begin
    Result := GHURCurrent.Level;

    If Result < Min Then
      ghAbort (ErrorCaption, ErrorText);
  End;

  Function ghCheckUserLevel (Const Min :Integer; Const ErrorText :String)
    :Integer; Overload;
  Begin
    Result := ghCheckUserLevel (Min, '', ErrorText);
  End;

  Procedure ghClearParams (Const Params :TParams);
  Var
    I :Integer;
  Begin
    For I := 0 To Params.Count - 1 Do
      Params [I].Clear;
  End;

  Function ghCloneDataSet (Const DataSet :TDataSet;
    Const AClass :TDataSetClass; Const Owner :TComponent = Nil;
    ExcludedProps :String = 'Active'; Const IncludeFields :Boolean = True;
    Const ExcludedFieldProps :String = '') :TDataSet; Overload;
  Var
    Params :TObject;
  Begin
    { We don't copy the IndexName property when the IndexFieldNames
      property has a value (otherwise it would be lost). }
    If ghPropStr (DataSet, 'IndexFieldNames') <> '' Then
      ExcludedProps := ghConcat (ExcludedProps, 'IndexName', ';');

    Result := TDataSet (ghClone (DataSet, AClass, Owner, ExcludedProps));

    If IncludeFields Then
      ghCopyFields (DataSet, Result, ExcludedFieldProps);

    Params := ghPropObj (DataSet, TParams);

    If Params <> Nil Then
      TParams (ghPropObj (Result, TParams)).Assign (TParams (Params));
  End;

  Function ghCloneDataSet (Const DataSet :TDataSet;
    Const Owner :TComponent = Nil; Const ExcludedProps :String = 'Active';
    Const IncludeFields :Boolean = True;
    Const ExcludedFieldProps :String = '') :TDataSet; Overload;
  Begin
    Result := ghCloneDataSet (DataSet, TDataSetClass (DataSet.ClassType),
      Owner, ExcludedProps, IncludeFields, ExcludedFieldProps);
  End;

  Function ghContainer (Const Control :TControl;
    Const AClass :TWinControlClass) :TWinControl; Overload;
  Begin
    If Control Is AClass Then
      Result := TWinControl (Control)
    Else
    Begin
      Result := Control.Parent;

      While Not ghIsOrNil (Result, AClass) Do
        Result := Result.Parent;
    End;
  End;

  Function ghContainer (Const AClass :TWinControlClass) :TWinControl;
    Overload;
  Begin
    If Screen.ActiveControl <> Nil Then
      Result := ghContainer (Screen.ActiveControl, AClass)
    Else
      Result := Nil;
  End;

  Procedure ghCopyDataSetProps (Const Source, Dest :TDataSet;
    Const Excludeds :String = ''; Const IncludeFields :Boolean = True;
    Const ExcludedFieldProps :String = '');
  Begin
    ghCopyProps (Source, Dest, Excludeds);

    If IncludeFields Then
      ghCopyFields (Source, Dest, ExcludedFieldProps);
  End;

  Procedure ghCopyFields (Const Source, Dest :TDataSet;
    Const ExcludedProps :String = '');
  Var
    I :Integer;
  Begin
    With Source Do
    Begin
      For I := 0 To Fields.Count - 1 Do
        TField (ghClone (Fields [I], Dest, ExcludedProps)).DataSet := Dest;

      For I := 0 To AggFields.Count - 1 Do
        TField (ghClone (AggFields [I], Dest, ExcludedProps)).DataSet :=
          Dest;
    End;
  End;

  Function ghDataSnapAttr (Const UpdateStatus :TUpdateStatus) :Integer;
  Begin
    Case UpdateStatus Of
      usModified : Result := dsRecModified;
      usInserted : Result := dsRecNew;
      usDeleted  : Result := dsRecDeleted;
      Else
        Result := 0;
    End;
  End;

  Function ghDataSnapAttrs (Const UpdateStatus :TUpdateStatusSet) :Integer;
  Var
    Status :TUpdateStatus;
  Begin
    Result := 0;

    For Status := Low (TUpdateStatus) To High (TUpdateStatus) Do
      If Status In UpdateStatus Then
        Result := Result Or ghDataSnapAttr (Status);
  End;

  Function ghDataSnapUpdateStatus (Const Attr :Integer) :TUpdateStatus;
  Begin
    Case Attr Of
      dsRecDeleted  : Result := usDeleted;
      dsRecNew      : Result := usInserted;
      dsRecModified : Result := usModified;
      Else
        Result := usUnmodified;
    End;
  End;

  Function ghDataSnapUpdateStatusSet (Const Attrs :Integer)
    :TUpdateStatusSet;
  Begin
    Result := [];

    If ghHasBit (Attrs, dsRecDeleted) Then
      Include (Result, usDeleted);

    If ghHasBit (Attrs, dsRecNew) Then
      Include (Result, usInserted);

    If ghHasBit (Attrs, dsRecModified) Then
      Include (Result, usModified);
  End;

  Function ghDestroying (Const Control :TControl) :Boolean;
  Begin
    Result := (csDestroying In Control.ComponentState) Or
      (csDestroyingHandle In Control.ControlState);
  End;

  Function ghDialogType (Const IconFlag :Integer) :TMsgDlgType;
  Begin
    Case IconFlag Of
      mb_IconStop        : Result := mtError;
      mb_IconQuestion    : Result := mtConfirmation;
      mb_IconWarning     : Result := mtWarning;
      mb_IconInformation : Result := mtInformation;
      Else
        Result := mtCustom;
    End;
  End;

  Function ghEqualFields (Const List1, List2 :TList) :Boolean;
  Var
    I :Integer;
  Begin
    Result := List1.Count = List2.Count;

    If Result Then
      For I := 0 To List1.Count - 1 Do
        If TField (List1 [I]).Value <> TField (List2 [I]).Value Then
        Begin
          Result := False;
          Exit;
        End;
  End;

  Function ghFieldCount (Const Fields :String) :Integer;
  Begin
    If Fields <> '' Then
      Result := ghCountTo (Fields, ';', Length (Fields) - 1) + 1
    Else
      Result := 0;
  End;

  Function ghFieldType (Const Name :String) :TFieldType;
  Begin
    Result := TFieldType (Max (ghFieldTypeInt (Name), 0 { ftUnknown }));
  End;

  Function ghFieldTypeEx (Const AType :String) :TFieldType;
  Var
    Value :Integer;
  Begin
    Result := ftUnknown;

    If AType = '' Then
      Exit;

    Value := ghFieldTypeInt (AType);

    If Value = -1 Then
      If ghEquals (AType, 'Double') Then
      Begin
        Result := ftFloat;
        Exit;
      End
      Else
      Begin
        Value := ghFieldTypeInt ('ft' + AType);

        If (Value = -1) And (AType [1] In ['T', 't']) Then
          Value := ghFieldTypeInt ('ft' + ghCopy (AType, 2));
      End;

    If Value > -1 Then
      Result := TFieldType (Value);
  End;

  Function ghFieldTypeInt (Const Name :String) :Integer;
  Begin
    Result := GetEnumValue (TypeInfo (TFieldType), Name);
  End;

  Function ghFiltering (Const DataSet :TDataSet) :TghDataSetFiltering;
  Begin
    With DataSet Do
    Begin
      Result.Active := Filtered;
      Result.Expr := Filter;
      Result.Event := OnFilterRecord;
      Result.Bookmark := GetBookmark;
    End;
  End;

  Function ghFocus (Const Control :TWinControl) :Boolean;
  Var
    Form :TCustomForm;
  Begin
    If Not ghCanFocus (Control, @Form) Then
      // Control is focusable but its container form is invisible/disabled
      If Form <> Nil Then
        ghMakeAble (Form)
      Else
      Begin
        Result := False;
        Exit;
      End;

    Control.SetFocus;
    Result := True;
  End;

  Function ghInform (Const Caption, Text :String;
    Const Options :Integer = 0) :Integer; Overload;
  Begin
    Result := ghShowMsg (Caption, Text,
      ghFlagIcon (Options, mb_IconInformation));
  End;

  Function ghInform (Const Text :String; Const Options :Integer = 0)
    :Integer; Overload;
  Begin
    Result := ghInform ('', Text, Options);
  End;

  {$Warn No_RetVal Off}
  Function ghInputParam (Const Params :TParams; Index :Integer = 0)
    :TParam;
  Var
    I :Integer;
  Begin
    For I := 0 To MaxInt Do
      If Params [I].ParamType In [ptInput, ptInputOutput] Then
      Begin
        Dec (Index);

        If Index = -1 Then
        Begin
          Result := Params [I];
          Exit;
        End;
      End;
  End;
  {$Warn No_RetVal On}

  Function ghInputParamCount (Const Params :TParams) :Integer;
  Var
    I :Integer;
  Begin
    Result := 0;

    For I := 0 To Params.Count - 1 Do
      If Params [I].ParamType In [ptInput, ptInputOutput] Then
        Inc (Result);
  End;

  Function ghIs (Const Window :THandle; Const AClass :TWinControlClass)
    :Boolean;
  Begin
    Result := FindControl (Window) Is AClass;
  End;

  Function ghIsButton (Const Window :THandle) :Boolean;
  Begin
    // To determinate when a Windows control is a push button

    Result := ghIs (Window, TButton) Or

      { NOTE: Some third party controls do not respond to the wm_GetDlgCode
        message properly. }
      ghGetDialogCodeButton (Window) Or

      ghClassButton (Window);
  End;

  Procedure ghMakeAble (Const Control :TControl);
  Begin
    Control.Enabled := True;
    Control.Visible := True;
  End;

    Type
      TDataSourceAccess = Class (TDataSource);
      TMasterDataLinkAccess = Class (TMasterDataLink);
  Function ghMasterLink (Const ADataSet :TDataSet) :TMasterDataLink;
  Var
    I :Integer;
  Begin
    { In some data set classes, the master data link is a private field
      without access property.  This function returns that data link. }

    If ADataSet.DataSource <> Nil Then
      With TDataSourceAccess (ADataSet.DataSource) Do
        For I := 0 To DataLinks.Count - 1 Do
          If (TDataLink (DataLinks [I]) Is TMasterDataLink) And
          (TMasterDataLinkAccess (DataLinks [I]).GetDetailDataSet =
          ADataSet) Then
          Begin
            // DataSet's master data link found
            Result := DataLinks [I];
            Exit;
          End;

    Result := Nil;
  End;

  Function ghModalLevel :Integer;
  Var
    I :Integer;
  Begin
    Result := 0;

    For I := 0 To Screen.CustomFormCount - 1 Do
      If fsModal In Screen.CustomForms [I].FormState Then
        Inc (Result);
  End;

  Function ghMsgIcon (Const Options :Integer; Var Icon :String) :Integer;
  Begin
    Result := Options And mb_IconMask;  // Icon flag
    Icon := ghMsgIconName (Result);  // Icon name
  End;

  Function ghMsgIconCheck (Const Options :Integer; Var Icon :String)
    :Integer;
  Begin
    Result := Options And mb_IconMask;  // Icon flag

    // Icon name
    If Icon = '' Then
      Icon := ghMsgIconName (Result);
  End;

  Function ghMsgIconName (Const Flag :Integer;
    Const ValidateFlag :Boolean = True) :String;
  Begin
    If Not ValidateFlag Or ghIsIconFlag (Flag) Then
      Result := GHMsgIconNames [ghDialogType (Flag)]
    Else
      Result := '';
  End;

  Function ghMsgBoxParams (Const Options :Integer = 0) :TMsgBoxParams;
    Overload;
  Begin
    Result.cbSize := ghZero (@Result, SizeOf (Result));
    Result.hwndOwner := Application.Handle;
    Result.dwStyle := Options;
  End;

  Function ghMsgBoxParams (Const Caption, Text :String;
    Const Options :Integer = 0) :TMsgBoxParams; Overload;
  Begin
    Result := ghMsgBoxParams (Options);
    Result.lpszText := PChar (Text);
    Result.lpszCaption := ghTitlePChr (Caption);
  End;

  Function ghMsgBoxParams (Const Text :String; Const Options :Integer = 0)
    :TMsgBoxParams; Overload;
  Begin
    Result := ghMsgBoxParams ('', Text, Options);
  End;

  Function ghMsgBoxParams (Const Caption, Text :String;
    Const Options :Integer; Const Icon :PChar;
    Const IconModule :THandle = 0) :TMsgBoxParams; Overload;
  Begin
    Result := ghMsgBoxParams (Caption, Text, Options);
    ghSetIcon (Result, Icon, IconModule)
  End;

  Function ghMsgBoxParams (Const Text :String; Const Options :Integer;
    Const Icon :PChar; Const IconModule :THandle = 0) :TMsgBoxParams;
    Overload;
  Begin
    Result := ghMsgBoxParams ('', Text, Options, Icon, IconModule);
  End;

  Function ghMsgBoxParams (Const Caption, Text :String;
    Const Icon :PChar; Const IconModule :THandle = 0) :TMsgBoxParams;
    Overload;
  Begin
    Result := ghMsgBoxParams (Caption, Text, 0, Icon, IconModule);
  End;

  Function ghMsgBoxParams (Const Options :Integer; Const Icon :PChar;
    Const ResourceModule :THandle = 0) :TMsgBoxParams; Overload;
  Begin
    Result := ghMsgBoxParams (Options);
    ghSetIcon (Result, Icon, ResourceModule)
  End;
  
  Function ghMsgBoxParamsMarks (Const Caption, Text :String;
    Const Options :Integer) :TMsgBoxParams;
  Begin
    Result := ghMsgBoxParams (Caption, Text, Options);

    If Options And mb_IconMask = 0 Then
      Result.dwStyle := Integer (Result.dwStyle) Or
        ghMsgIconFlagStrs ([Caption, Text]);  // Icon flag from marks (!/?)
  End;

  {$Warn No_RetVal Off}
  Function ghOutputParam (Const Params :TParams; Index :Integer = 0)
    :TParam;
  Var
    I :Integer;
  Begin
    For I := 0 To MaxInt Do
      If Params [I].ParamType In [ptOutput, ptInputOutput, ptResult] Then
      Begin
        Dec (Index);

        If Index = -1 Then
        Begin
          Result := Params [I];
          Exit;
        End;
      End;
  End;
  {$Warn No_RetVal On}

  Function ghOutputParamCount (Const Params :TParams) :Integer;
  Var
    I :Integer;
  Begin
    Result := 0;

    For I := 0 To Params.Count - 1 Do
      If Params [I].ParamType In [ptOutput, ptInputOutput, ptResult] Then
        Inc (Result);
  End;

  Function ghPressChr (Const Control :TControl; Const Value :Char;
    Const KeyData :Integer = 0) :Integer;
  Begin
    Result := Control.Perform (wm_Char, Ord (Value), KeyData);
  End;

  Function ghRootControl (Const Control :TControl) :TControl;
  Begin
    Result := Control;

    While (Result.Parent <> Nil) And Not (Result.Parent Is TCustomForm) Do
      Result := Result.Parent;
  End;

  Function ghSetCursor (Const Value :TCursor) :TCursor;
  Begin
    Result := Screen.Cursor;
    Screen.Cursor := Value;
  End;

  Function ghShowMsg (Const Params :TMsgBoxParams) :Integer; Overload;
  Var
    WindowList :Pointer;
  Begin
    WindowList := DisableTaskWindows (0);

    Try
      Result := Integer (MessageBoxIndirect (Params));
    Finally
      EnableTaskWindows (WindowList);
    End;
  End;

  Function ghShowMsg (Const Params :TMsgBoxParams; Const Sound :Integer)
    :Integer; Overload;
  Begin
    MessageBeep (Sound);
    Result := ghShowMsg (Params);
  End;

  Function ghShowMsg (Const Caption, Text :String;
    Const Options :Integer = 0; Icon :String = '') :Integer; Overload;
  Var
    IconFlag :Integer Absolute Result;
    Params :TMsgBoxParams;
  Begin
    Params := ghMsgBoxParamsMarks (Caption, Text, Options);
    IconFlag := ghMsgIconCheck (Params.dwStyle, Icon);

    If (Icon <> '') And (ghFindIcon (Icon) <> 0) Then
    Begin
      ghSetIcon (Params, ghPlatStr (Icon).Ptr);

      If IconFlag <> 0 Then
        MessageBeep (IconFlag);
    End;

    Result := ghShowMsg (Params);
  End;

  Function ghShowMsg (Const Text :String; Const Options :Integer = 0;
    Const Icon :String = '') :Integer; Overload;
  Begin
    Result := ghShowMsg ('', Text, Options, Icon);
  End;

  Function ghShowMsg (Const Caption, Text, Icon :String) :Integer;
    Overload;
  Begin
    Result := ghShowMsg (Caption, Text, 0, Icon);
  End;

  Function ghShowMsgIcon (Const Text, Icon :String) :Integer; Overload;
  Begin
    Result := ghShowMsg (Text, 0, Icon);
  End;

  Function ghShowStop (Const Caption, Text :String;
    Const Options :Integer = 0; Const FocusControl :TWinControl = Nil)
    :Integer; Overload;
  Begin
    If FocusControl <> Nil Then
      ghFocus (FocusControl);

    Result := ghShowMsg (Caption, Text, ghFlagIcon (Options, mb_IconStop));
  End;

  Function ghShowStop (Const Text :String; Const Options :Integer = 0;
    Const FocusControl :TWinControl = Nil) :Integer; Overload;
  Begin
    Result := ghShowStop ('', Text, Options, FocusControl);
  End;

  Function ghShowStop (Const Caption, Text :String;
    Const FocusControl :TWinControl) :Integer; Overload;
  Begin
    Result := ghShowStop (Caption, Text, 0, FocusControl);
  End;

  Function ghShowStop (Const Text :String; Const FocusControl :TWinControl)
    :Integer; Overload;
  Begin
    Result := ghShowStop (Text, 0, FocusControl);
  End;

  Function ghSQL (Const DataSet :TDataSet) :String;
  Var
    SQL :TStrings;
  Begin
    SQL := ghSQLList (DataSet);

    If SQL <> Nil Then
      Result := SQL.Text
    Else
      Result := '';
  End;

  Function ghSQLList (Const DataSet :TDataSet) :TStrings;
  Begin
    Result := TStrings (ghPropObj (DataSet, 'SQL', TStrings));
  End;

  Function ghSQLTable (Const DataSet :TDataSet) :String;
  Begin
    Result := ghSQL (DataSet);

    If Result <> '' Then
      Result := GetTableNameFromSQL (Result)
  End;

  Function ghTable (Const DataSet :TDataSet) :String;
  Begin
    If IsPublishedProp (DataSet, 'TableName') Then
      Result := GetPropValue (DataSet, 'TableName', False)
    Else
      Result := ghSQLTable (DataSet)
  End;

  Function ghTitle (Const Value :String) :String;
  Begin
    If Value <> '' Then
      Result := Value
    Else
      Result := Application.Title;  // Application.Title by default
  End;

  Function ghTitlePChr :PChar; Overload;
  Begin
    AppTitle := Application.Title;
    Result := PChar (AppTitle);
  End;

  Function ghTitlePChr (Const Value :String) :PChar; Overload;
  Begin
    If Value <> '' Then
      Result := Pointer (Value)
    Else
      Result := ghTitlePChr;  // Application.Title by default
  End;

  Function ghTitlePtr :Pointer; Overload;
  Begin
    AppTitle := Application.Title;
    Result := Pointer (AppTitle);
  End;

  Function ghTitlePtr (Const Value :String) :Pointer; Overload;
  Begin
    If Value <> '' Then
      Result := Pointer (Value)
    Else
      Result := ghTitlePtr;  // Application.Title by default
  End;

  Function ghUsable (Const Control :TControl) :Boolean;
  Begin
    Result := ghAble (Control) And Not ghDestroying (Control);
  End;

  Function ghWantEnter (Const Window :THandle) :Boolean;
  Begin
    Result := ghWantAllKeys (Window) Or ghIsButton (Window);
  End;

  Function ghWarn (Const Caption, Text :String;
    Const Options :Integer = 0) :Integer; Overload;
  Begin
    Result := ghShowMsg (Caption, Text,
      ghFlagIcon (Options, mb_IconWarning));
  End;

  Function ghWarn (Const Text :String; Const Options :Integer = 0)
    :Integer; Overload;
  Begin
    Result := ghWarn ('', Text, Options);
  End;

End.

